# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This is a set of regression tests for vo.
"""

# STDLIB
import difflib
import gzip
import io
import pathlib
import re
import sys
from unittest import mock

# THIRD-PARTY
import numpy as np
import pytest
from numpy.testing import assert_array_equal

# LOCAL
from astropy.io.votable import tree
from astropy.io.votable.exceptions import W39, VOTableSpecError, VOWarning
from astropy.io.votable.table import parse, parse_single_table, validate
from astropy.io.votable.xmlutil import validate_schema
from astropy.utils.data import get_pkg_data_filename, get_pkg_data_filenames

# Determine the kind of float formatting in this build of Python
if hasattr(sys, "float_repr_style"):
    legacy_float_repr = sys.float_repr_style == "legacy"
else:
    legacy_float_repr = sys.platform.startswith("win")


def assert_validate_schema(filename, version):
    if sys.platform.startswith("win"):
        return

    try:
        rc, stdout, stderr = validate_schema(filename, version)
    except OSError:
        # If xmllint is not installed, we want the test to pass anyway
        return
    assert rc == 0, "File did not validate against VOTable schema"


def test_parse_single_table():
    with np.errstate(over="ignore"):
        # https://github.com/astropy/astropy/issues/13341
        table = parse_single_table(get_pkg_data_filename("data/regression.xml"))
    assert isinstance(table, tree.TableElement)
    assert len(table.array) == 5


def test_parse_single_table2():
    with np.errstate(over="ignore"):
        # https://github.com/astropy/astropy/issues/13341
        table2 = parse_single_table(
            get_pkg_data_filename("data/regression.xml"), table_number=1
        )
    assert isinstance(table2, tree.TableElement)
    assert len(table2.array) == 1
    assert len(table2.array.dtype.names) == 28


def test_parse_single_table3():
    with pytest.raises(IndexError):
        parse_single_table(get_pkg_data_filename("data/regression.xml"), table_number=3)


def _test_regression(tmp_path, _python_based=False, binary_mode=1):
    # Read the VOTABLE
    votable = parse(
        get_pkg_data_filename("data/regression.xml"),
        _debug_python_based_parser=_python_based,
    )
    table = votable.get_first_table()

    dtypes = [
        (("string test", "string_test"), "|O8"),
        (("fixed string test", "string_test_2"), "<U10"),
        ("unicode_test", "|O8"),
        (("unicode test", "fixed_unicode_test"), "<U10"),
        (("string array test", "string_array_test"), "<U4"),
        ("unsignedByte", "|u1"),
        ("short", "<i2"),
        ("int", "<i4"),
        ("long", "<i8"),
        ("double", "<f8"),
        ("float", "<f4"),
        ("array", "|O8"),
        ("bit", "|b1"),
        ("bitarray", "|b1", (3, 2)),
        ("bitvararray", "|O8"),
        ("bitvararray2", "|O8"),
        ("floatComplex", "<c8"),
        ("doubleComplex", "<c16"),
        ("doubleComplexArray", "|O8"),
        ("doubleComplexArrayFixed", "<c16", (2,)),
        ("boolean", "|b1"),
        ("booleanArray", "|b1", (4,)),
        ("nulls", "<i4"),
        ("nulls_array", "<i4", (2, 2)),
        ("precision1", "<f8"),
        ("precision2", "<f8"),
        ("doublearray", "|O8"),
        ("bitarray2", "|b1", (16,)),
    ]
    if sys.byteorder == "big":
        new_dtypes = []
        for dtype in dtypes:
            dtype = list(dtype)
            dtype[1] = dtype[1].replace("<", ">")
            new_dtypes.append(tuple(dtype))
        dtypes = new_dtypes
    assert table.array.dtype == dtypes

    votable.to_xml(
        str(tmp_path / "regression.tabledata.xml"),
        _debug_python_based_parser=_python_based,
    )
    assert_validate_schema(str(tmp_path / "regression.tabledata.xml"), votable.version)

    if binary_mode == 1:
        votable.get_first_table().format = "binary"
        votable.version = "1.1"
    elif binary_mode == 2:
        votable.get_first_table()._config["version_1_3_or_later"] = True
        votable.get_first_table().format = "binary2"
        votable.version = "1.3"

    # Also try passing a file handle
    with open(str(tmp_path / "regression.binary.xml"), "wb") as fd:
        votable.to_xml(fd, _debug_python_based_parser=_python_based)
    assert_validate_schema(str(tmp_path / "regression.binary.xml"), votable.version)
    # Also try passing a file handle
    with open(str(tmp_path / "regression.binary.xml"), "rb") as fd:
        votable2 = parse(fd, _debug_python_based_parser=_python_based)
    votable2.get_first_table().format = "tabledata"
    votable2.to_xml(
        str(tmp_path / "regression.bin.tabledata.xml"),
        _astropy_version="testing",
        _debug_python_based_parser=_python_based,
    )
    assert_validate_schema(
        str(tmp_path / "regression.bin.tabledata.xml"), votable.version
    )

    with open(
        get_pkg_data_filename(
            f"data/regression.bin.tabledata.truth.{votable.version}.xml"
        ),
        encoding="utf-8",
    ) as fd:
        truth = fd.readlines()
    with open(str(tmp_path / "regression.bin.tabledata.xml"), encoding="utf-8") as fd:
        output = fd.readlines()

    # If the lines happen to be different, print a diff
    # This is convenient for debugging
    sys.stdout.writelines(
        difflib.unified_diff(truth, output, fromfile="truth", tofile="output")
    )

    assert truth == output

    # Test implicit gzip saving
    votable2.to_xml(
        str(tmp_path / "regression.bin.tabledata.xml.gz"),
        _astropy_version="testing",
        _debug_python_based_parser=_python_based,
    )
    with gzip.GzipFile(str(tmp_path / "regression.bin.tabledata.xml.gz"), "rb") as gzfd:
        output = gzfd.readlines()
    output = [x.decode("utf-8").rstrip() for x in output]
    truth = [x.rstrip() for x in truth]

    assert truth == output


@pytest.mark.xfail("legacy_float_repr")
def test_regression(tmp_path):
    # W39: Bit values can not be masked
    with pytest.warns(W39), np.errstate(over="ignore"):
        _test_regression(tmp_path, False)


@pytest.mark.xfail("legacy_float_repr")
def test_regression_python_based_parser(tmp_path):
    # W39: Bit values can not be masked
    with pytest.warns(W39), np.errstate(over="ignore"):
        _test_regression(tmp_path, True)


@pytest.mark.xfail("legacy_float_repr")
def test_regression_binary2(tmp_path):
    # W39: Bit values can not be masked
    with pytest.warns(W39), np.errstate(over="ignore"):
        _test_regression(tmp_path, False, 2)


class TestFixups:
    def setup_class(self):
        with np.errstate(over="ignore"):
            # https://github.com/astropy/astropy/issues/13341
            self.table = parse(
                get_pkg_data_filename("data/regression.xml")
            ).get_first_table()
        self.array = self.table.array
        self.mask = self.table.array.mask

    def test_implicit_id(self):
        assert_array_equal(self.array["string_test_2"], self.array["fixed string test"])


class TestReferences:
    def setup_class(self):
        with np.errstate(over="ignore"):
            # https://github.com/astropy/astropy/issues/13341
            self.votable = parse(get_pkg_data_filename("data/regression.xml"))
        self.table = self.votable.get_first_table()
        self.array = self.table.array
        self.mask = self.table.array.mask

    def test_fieldref(self):
        fieldref = self.table.groups[1].entries[0]
        assert isinstance(fieldref, tree.FieldRef)
        assert fieldref.get_ref().name == "boolean"
        assert fieldref.get_ref().datatype == "boolean"

    def test_paramref(self):
        paramref = self.table.groups[0].entries[0]
        assert isinstance(paramref, tree.ParamRef)
        assert paramref.get_ref().name == "INPUT"
        assert paramref.get_ref().datatype == "float"

    def test_iter_fields_and_params_on_a_group(self):
        assert len(list(self.table.groups[1].iter_fields_and_params())) == 2

    def test_iter_groups_on_a_group(self):
        assert len(list(self.table.groups[1].iter_groups())) == 1

    def test_iter_groups(self):
        # Because of the ref'd table, there are more logical groups
        # than actually exist in the file
        assert len(list(self.votable.iter_groups())) == 9

    def test_ref_table(self):
        tables = list(self.votable.iter_tables())
        for x, y in zip(tables[0].array.data[0], tables[1].array.data[0]):
            assert_array_equal(x, y)

    def test_iter_coosys(self):
        assert len(list(self.votable.iter_coosys())) == 1


@pytest.mark.parametrize(
    "columns, expected_missing",
    [
        # a single non-existent column
        pytest.param(["c1"], ["c1"], id="basic"),
        # multiple missing columns (checking that order is preserved)
        pytest.param(["c1", "c2", "c3"], ["c1", "c2", "c3"], id="check-ordering"),
        # mixing existing with missing columns
        pytest.param(["c1", "string_test", "c2"], ["c1", "c2"], id="list-only-missing"),
    ],
)
def test_select_missing_columns_error_message(columns, expected_missing):
    # see https://github.com/astropy/astropy/pull/15956
    filename = get_pkg_data_filename("data/regression.xml")
    with pytest.raises(
        ValueError,
        match=re.escape(f"Columns {expected_missing!r} were not found in fields list"),
    ):
        parse_single_table(filename, columns=columns)


def test_select_columns_by_index():
    columns = [0, 5, 13]
    table = parse(
        get_pkg_data_filename("data/regression.xml"), columns=columns
    ).get_first_table()
    array = table.array
    mask = table.array.mask
    assert array["string_test"][0] == "String & test"
    columns = ["string_test", "unsignedByte", "bitarray"]
    for c in columns:
        assert not np.all(mask[c])
    assert np.all(mask["unicode_test"])


def test_select_columns_by_name():
    columns = ["string_test", "unsignedByte", "bitarray"]
    table = parse(
        get_pkg_data_filename("data/regression.xml"), columns=columns
    ).get_first_table()
    array = table.array
    mask = table.array.mask
    assert array["string_test"][0] == "String & test"
    for c in columns:
        assert not np.all(mask[c])
    assert np.all(mask["unicode_test"])


class TestParse:
    def setup_class(self):
        with np.errstate(over="ignore"):
            # https://github.com/astropy/astropy/issues/13341
            self.votable = parse(get_pkg_data_filename("data/regression.xml"))
        self.table = self.votable.get_first_table()
        self.array = self.table.array
        self.mask = self.table.array.mask

    def test_string_test(self):
        assert issubclass(self.array["string_test"].dtype.type, np.object_)
        assert_array_equal(
            self.array["string_test"],
            ["String & test", "String &amp; test", "XXXX", "", ""],
        )

    def test_fixed_string_test(self):
        assert issubclass(self.array["string_test_2"].dtype.type, np.str_)
        assert_array_equal(
            self.array["string_test_2"], ["Fixed stri", "0123456789", "XXXX", "", ""]
        )

    def test_unicode_test(self):
        assert issubclass(self.array["unicode_test"].dtype.type, np.object_)
        assert_array_equal(
            self.array["unicode_test"],
            ["Ceçi n'est pas un pipe", "வணக்கம்", "XXXX", "", ""],
        )

    def test_fixed_unicode_test(self):
        assert issubclass(self.array["fixed_unicode_test"].dtype.type, np.str_)
        assert_array_equal(
            self.array["fixed_unicode_test"],
            ["Ceçi n'est", "வணக்கம்", "0123456789", "", ""],
        )

    def test_unsignedByte(self):
        assert issubclass(self.array["unsignedByte"].dtype.type, np.uint8)
        assert_array_equal(self.array["unsignedByte"], [128, 255, 0, 255, 255])
        assert not np.any(self.mask["unsignedByte"])

    def test_short(self):
        assert issubclass(self.array["short"].dtype.type, np.int16)
        assert_array_equal(self.array["short"], [4096, 32767, -4096, 32767, 32767])
        assert not np.any(self.mask["short"])

    def test_int(self):
        assert issubclass(self.array["int"].dtype.type, np.int32)
        assert_array_equal(
            self.array["int"], [268435456, 2147483647, -268435456, 268435455, 123456789]
        )
        assert_array_equal(self.mask["int"], [False, False, False, False, True])

    def test_long(self):
        assert issubclass(self.array["long"].dtype.type, np.int64)
        assert_array_equal(
            self.array["long"],
            [
                922337203685477,
                123456789,
                -1152921504606846976,
                1152921504606846975,
                123456789,
            ],
        )
        assert_array_equal(self.mask["long"], [False, True, False, False, True])

    def test_double(self):
        assert issubclass(self.array["double"].dtype.type, np.float64)
        assert_array_equal(
            self.array["double"], [8.9990234375, 0.0, np.inf, np.nan, -np.inf]
        )
        assert_array_equal(self.mask["double"], [False, False, False, True, False])

    def test_float(self):
        assert issubclass(self.array["float"].dtype.type, np.float32)
        assert_array_equal(self.array["float"], [1.0, 0.0, np.inf, np.inf, np.nan])
        assert_array_equal(self.mask["float"], [False, False, False, False, True])

    def test_array(self):
        assert issubclass(self.array["array"].dtype.type, np.object_)
        match = [
            [],
            [[42, 32], [12, 32]],
            [[12, 34], [56, 78], [87, 65], [43, 21]],
            [[-1, 23]],
            [[31, -1]],
        ]
        for a, b in zip(self.array["array"], match):
            # assert issubclass(a.dtype.type, np.int64)
            # assert a.shape[1] == 2
            for a0, b0 in zip(a, b):
                assert issubclass(a0.dtype.type, np.int64)
                assert_array_equal(a0, b0)
        assert self.array.data["array"][3].mask[0][0]
        assert self.array.data["array"][4].mask[0][1]

    def test_bit(self):
        assert issubclass(self.array["bit"].dtype.type, np.bool_)
        assert_array_equal(self.array["bit"], [True, False, True, False, False])

    def test_bit_mask(self):
        assert_array_equal(self.mask["bit"], [False, False, False, False, True])

    def test_bitarray(self):
        assert issubclass(self.array["bitarray"].dtype.type, np.bool_)
        assert self.array["bitarray"].shape == (5, 3, 2)
        assert_array_equal(
            self.array["bitarray"],
            [
                [[True, False], [True, True], [False, True]],
                [[False, True], [False, False], [True, True]],
                [[True, True], [True, False], [False, False]],
                [[False, False], [False, False], [False, False]],
                [[False, False], [False, False], [False, False]],
            ],
        )

    def test_bitarray_mask(self):
        assert_array_equal(
            self.mask["bitarray"],
            [
                [[False, False], [False, False], [False, False]],
                [[False, False], [False, False], [False, False]],
                [[False, False], [False, False], [False, False]],
                [[True, True], [True, True], [True, True]],
                [[True, True], [True, True], [True, True]],
            ],
        )

    def test_bitvararray(self):
        assert issubclass(self.array["bitvararray"].dtype.type, np.object_)
        match = [
            [True, True, True],
            [False, False, False, False, False],
            [True, False, True, False, True],
            [],
            [],
        ]
        for a, b in zip(self.array["bitvararray"], match):
            assert_array_equal(a, b)
        match_mask = [
            [False, False, False],
            [False, False, False, False, False],
            [False, False, False, False, False],
            False,
            False,
        ]
        for a, b in zip(self.array["bitvararray"], match_mask):
            assert_array_equal(a.mask, b)

    def test_bitvararray2(self):
        assert issubclass(self.array["bitvararray2"].dtype.type, np.object_)
        match = [
            [],
            [
                [[False, True], [False, False], [True, False]],
                [[True, False], [True, False], [True, False]],
            ],
            [[[True, True], [True, True], [True, True]]],
            [],
            [],
        ]
        for a, b in zip(self.array["bitvararray2"], match):
            for a0, b0 in zip(a, b):
                assert a0.shape == (3, 2)
                assert issubclass(a0.dtype.type, np.bool_)
                assert_array_equal(a0, b0)

    def test_floatComplex(self):
        assert issubclass(self.array["floatComplex"].dtype.type, np.complex64)
        assert_array_equal(
            self.array["floatComplex"],
            [np.nan + 0j, 0 + 0j, 0 + -1j, np.nan + 0j, np.nan + 0j],
        )
        assert_array_equal(self.mask["floatComplex"], [True, False, False, True, True])

    def test_doubleComplex(self):
        assert issubclass(self.array["doubleComplex"].dtype.type, np.complex128)
        assert_array_equal(
            self.array["doubleComplex"],
            [np.nan + 0j, 0 + 0j, 0 + -1j, np.nan + (np.inf * 1j), np.nan + 0j],
        )
        assert_array_equal(self.mask["doubleComplex"], [True, False, False, True, True])

    def test_doubleComplexArray(self):
        assert issubclass(self.array["doubleComplexArray"].dtype.type, np.object_)
        assert [len(x) for x in self.array["doubleComplexArray"]] == [0, 2, 2, 0, 0]

    def test_boolean(self):
        assert issubclass(self.array["boolean"].dtype.type, np.bool_)
        assert_array_equal(self.array["boolean"], [True, False, True, False, False])

    def test_boolean_mask(self):
        assert_array_equal(self.mask["boolean"], [False, False, False, False, True])

    def test_boolean_array(self):
        assert issubclass(self.array["booleanArray"].dtype.type, np.bool_)
        assert_array_equal(
            self.array["booleanArray"],
            [
                [True, True, True, True],
                [True, True, False, True],
                [True, True, False, True],
                [False, False, False, False],
                [False, False, False, False],
            ],
        )

    def test_boolean_array_mask(self):
        assert_array_equal(
            self.mask["booleanArray"],
            [
                [False, False, False, False],
                [False, False, False, False],
                [False, False, True, False],
                [True, True, True, True],
                [True, True, True, True],
            ],
        )

    def test_nulls(self):
        assert_array_equal(self.array["nulls"], [0, -9, 2, -9, -9])
        assert_array_equal(self.mask["nulls"], [False, True, False, True, True])

    def test_nulls_array(self):
        assert_array_equal(
            self.array["nulls_array"],
            [
                [[-9, -9], [-9, -9]],
                [[0, 1], [2, 3]],
                [[-9, 0], [-9, 1]],
                [[0, -9], [1, -9]],
                [[-9, -9], [-9, -9]],
            ],
        )
        assert_array_equal(
            self.mask["nulls_array"],
            [
                [[True, True], [True, True]],
                [[False, False], [False, False]],
                [[True, False], [True, False]],
                [[False, True], [False, True]],
                [[True, True], [True, True]],
            ],
        )

    def test_double_array(self):
        assert issubclass(self.array["doublearray"].dtype.type, np.object_)
        assert len(self.array["doublearray"][0]) == 0
        assert_array_equal(
            self.array["doublearray"][1], [0, 1, np.inf, -np.inf, np.nan, 0, -1]
        )
        assert_array_equal(
            self.array.data["doublearray"][1].mask,
            [False, False, False, False, False, False, True],
        )

    def test_bit_array2(self):
        assert_array_equal(
            self.array["bitarray2"][0],
            [
                True,
                True,
                True,
                True,
                False,
                False,
                False,
                False,
                True,
                True,
                True,
                True,
                False,
                False,
                False,
                False,
            ],
        )

    def test_bit_array2_mask(self):
        assert not np.any(self.mask["bitarray2"][0])
        assert np.all(self.mask["bitarray2"][1:])

    def test_get_coosys_by_id(self):
        coosys = self.votable.get_coosys_by_id("J2000")
        assert coosys.system == "eq_FK5"

    def test_get_field_by_utype(self):
        fields = list(self.votable.get_fields_by_utype("myint"))
        assert fields[0].name == "int"
        assert fields[0].values.min == -1000

    def test_get_info_by_id(self):
        info = self.votable.get_info_by_id("QUERY_STATUS")
        assert info.value == "OK"

        if self.votable.version != "1.1":
            info = self.votable.get_info_by_id("ErrorInfo")
            assert info.value == "One might expect to find some INFO here, too..."

    def test_repr(self):
        assert "3 tables" in repr(self.votable)
        assert (
            repr(list(self.votable.iter_fields_and_params())[0])
            == '<PARAM ID="awesome" arraysize="*" datatype="float" '
            'name="INPUT" unit="deg" value="[0.0 0.0]"/>'
        )
        # Smoke test
        repr(list(self.votable.iter_groups()))

        # Resource
        assert repr(self.votable.resources) == "[</>]"

        # Table
        assert repr(self.table).startswith("<VOTable")


class TestThroughTableData(TestParse):
    def setup_class(self):
        with np.errstate(over="ignore"):
            # https://github.com/astropy/astropy/issues/13341
            votable = parse(get_pkg_data_filename("data/regression.xml"))

        self.xmlout = bio = io.BytesIO()
        # W39: Bit values can not be masked
        with pytest.warns(W39):
            votable.to_xml(bio)
        bio.seek(0)
        self.votable = parse(bio)
        self.table = self.votable.get_first_table()
        self.array = self.table.array
        self.mask = self.table.array.mask

    def test_bit_mask(self):
        assert_array_equal(self.mask["bit"], [False, False, False, False, False])

    def test_bitarray_mask(self):
        assert not np.any(self.mask["bitarray"])

    def test_bit_array2_mask(self):
        assert not np.any(self.mask["bitarray2"])

    def test_schema(self, tmp_path):
        # have to use an actual file because assert_validate_schema only works
        # on filenames, not file-like objects
        fn = tmp_path / "test_through_tabledata.xml"
        with open(fn, "wb") as f:
            f.write(self.xmlout.getvalue())
        assert_validate_schema(fn, "1.1")


class TestThroughBinary(TestParse):
    def setup_class(self):
        with np.errstate(over="ignore"):
            # https://github.com/astropy/astropy/issues/13341
            votable = parse(get_pkg_data_filename("data/regression.xml"))
        votable.get_first_table().format = "binary"

        self.xmlout = bio = io.BytesIO()
        # W39: Bit values can not be masked
        with pytest.warns(W39):
            votable.to_xml(bio)
        bio.seek(0)
        self.votable = parse(bio)

        self.table = self.votable.get_first_table()
        self.array = self.table.array
        self.mask = self.table.array.mask

    # Masked values in bit fields don't roundtrip through the binary
    # representation -- that's not a bug, just a limitation, so
    # override the mask array checks here.
    def test_bit_mask(self):
        assert not np.any(self.mask["bit"])

    def test_bitarray_mask(self):
        assert not np.any(self.mask["bitarray"])

    def test_bit_array2_mask(self):
        assert not np.any(self.mask["bitarray2"])


class TestThroughBinary2(TestParse):
    def setup_class(self):
        with np.errstate(over="ignore"):
            # https://github.com/astropy/astropy/issues/13341
            votable = parse(get_pkg_data_filename("data/regression.xml"))
        votable.version = "1.3"
        votable.get_first_table()._config["version_1_3_or_later"] = True
        votable.get_first_table().format = "binary2"

        self.xmlout = bio = io.BytesIO()
        # W39: Bit values can not be masked
        with pytest.warns(W39):
            votable.to_xml(bio)
        bio.seek(0)
        self.votable = parse(bio)

        self.table = self.votable.get_first_table()
        self.array = self.table.array
        self.mask = self.table.array.mask

    def test_get_coosys_by_id(self):
        # No COOSYS in VOTable 1.2 or later
        pass


def table_from_scratch():
    from astropy.io.votable.tree import Field, Resource, TableElement, VOTableFile

    # Create a new VOTable file...
    votable = VOTableFile()

    # ...with one resource...
    resource = Resource()
    votable.resources.append(resource)

    # ... with one table
    table = TableElement(votable)
    resource.tables.append(table)

    # Define some fields
    table.fields.extend(
        [
            Field(votable, ID="filename", datatype="char"),
            Field(votable, ID="matrix", datatype="double", arraysize="2x2"),
        ]
    )

    # Now, use those field definitions to create the numpy record arrays, with
    # the given number of rows
    table.create_arrays(2)

    # Now table.array can be filled with data
    table.array[0] = ("test1.xml", [[1, 0], [0, 1]])
    table.array[1] = ("test2.xml", [[0.5, 0.3], [0.2, 0.1]])

    # Now write the whole thing to a file.
    # Note, we have to use the top-level votable file object
    out = io.StringIO()
    votable.to_xml(out)


# https://github.com/astropy/astropy/issues/13341
@np.errstate(over="ignore")
def test_open_files():
    for filename in get_pkg_data_filenames("data", pattern="*.xml"):
        if not filename.endswith(
            ("custom_datatype.xml", "timesys_errors.xml", "parquet_binary.xml")
        ):
            parse(filename)


def test_too_many_columns():
    with pytest.raises(VOTableSpecError):
        parse(get_pkg_data_filename("data/too_many_columns.xml.gz"))


def test_build_from_scratch(tmp_path):
    # Create a new VOTable file...
    votable = tree.VOTableFile()

    # ...with one resource...
    resource = tree.Resource()
    votable.resources.append(resource)

    # ... with one table
    table = tree.TableElement(votable)
    resource.tables.append(table)

    # Define some fields
    table.fields.extend(
        [
            tree.Field(
                votable, ID="filename", name="filename", datatype="char", arraysize="1"
            ),
            tree.Field(
                votable, ID="matrix", name="matrix", datatype="double", arraysize="2x2"
            ),
        ]
    )

    # Now, use those field definitions to create the numpy record arrays, with
    # the given number of rows
    table.create_arrays(2)

    # Now table.array can be filled with data
    table.array[0] = ("test1.xml", [[1, 0], [0, 1]])
    table.array[1] = ("test2.xml", [[0.5, 0.3], [0.2, 0.1]])

    # Now write the whole thing to a file.
    # Note, we have to use the top-level votable file object
    votable.to_xml(str(tmp_path / "new_votable.xml"))

    votable = parse(str(tmp_path / "new_votable.xml"))

    table = votable.get_first_table()
    assert_array_equal(
        table.array.mask,
        np.array(
            [
                (False, [[False, False], [False, False]]),
                (False, [[False, False], [False, False]]),
            ],
            dtype=[("filename", "?"), ("matrix", "?", (2, 2))],
        ),
    )


def test_validate(test_path_object=False):
    """
    test_path_object is needed for test below ``test_validate_path_object``
    so that file could be passed as pathlib.Path object.
    """
    output = io.StringIO()
    fpath = get_pkg_data_filename("data/regression.xml")
    if test_path_object:
        fpath = pathlib.Path(fpath)

    # We can't test xmllint, because we can't rely on it being on the
    # user's machine.
    result = validate(fpath, output, xmllint=False)

    assert result is False

    output.seek(0)
    output = output.readlines()

    # Uncomment to generate new groundtruth
    # with open('validation.txt', 'wt', encoding='utf-8') as fd:
    #    fd.write(u''.join(output))

    with open(get_pkg_data_filename("data/validation.txt"), encoding="utf-8") as fd:
        truth = fd.readlines()

    truth = truth[1:]
    output = output[1:-1]

    sys.stdout.writelines(
        difflib.unified_diff(truth, output, fromfile="truth", tofile="output")
    )

    assert truth == output


@mock.patch("subprocess.Popen")
def test_validate_xmllint_true(mock_subproc_popen):
    process_mock = mock.Mock()
    attrs = {"communicate.return_value": ("ok", "ko"), "returncode": 0}
    process_mock.configure_mock(**attrs)
    mock_subproc_popen.return_value = process_mock

    assert validate(get_pkg_data_filename("data/empty_table.xml"), xmllint=True)


def test_validate_path_object():
    """Validating when source is passed as path object (#4412)."""
    test_validate(test_path_object=True)


def test_gzip_filehandles(tmp_path):
    with np.errstate(over="ignore"):
        # https://github.com/astropy/astropy/issues/13341
        votable = parse(get_pkg_data_filename("data/regression.xml"))

    # W39: Bit values can not be masked
    with pytest.warns(W39):
        with open(tmp_path / "regression.compressed.xml", "wb") as fd:
            votable.to_xml(fd, compressed=True, _astropy_version="testing")

    with open(tmp_path / "regression.compressed.xml", "rb") as fd:
        votable = parse(fd)


def test_from_scratch_example():
    _run_test_from_scratch_example()


def _run_test_from_scratch_example():
    from astropy.io.votable.tree import Field, Resource, TableElement, VOTableFile

    # Create a new VOTable file...
    votable = VOTableFile()

    # ...with one resource...
    resource = Resource()
    votable.resources.append(resource)

    # ... with one table
    table = TableElement(votable)
    resource.tables.append(table)

    # Define some fields
    table.fields.extend(
        [
            Field(votable, name="filename", datatype="char", arraysize="*"),
            Field(votable, name="matrix", datatype="double", arraysize="2x2"),
        ]
    )

    # Now, use those field definitions to create the numpy record arrays, with
    # the given number of rows
    table.create_arrays(2)

    # Now table.array can be filled with data
    table.array[0] = ("test1.xml", [[1, 0], [0, 1]])
    table.array[1] = ("test2.xml", [[0.5, 0.3], [0.2, 0.1]])

    assert table.array[0][0] == "test1.xml"


def test_fileobj():
    # Assert that what we get back is a raw C file pointer
    # so it will be super fast in the C extension.
    from astropy.utils.xml import iterparser

    filename = get_pkg_data_filename("data/regression.xml")
    with iterparser._convert_to_fd_or_read_function(filename) as fd:
        if sys.platform == "win32":
            fd()
        else:
            assert isinstance(fd, io.FileIO)


def test_nonstandard_units():
    from astropy import units as u

    votable = parse(get_pkg_data_filename("data/nonstandard_units.xml"))

    assert isinstance(votable.get_first_table().fields[0].unit, u.UnrecognizedUnit)

    votable = parse(
        get_pkg_data_filename("data/nonstandard_units.xml"), unit_format="generic"
    )

    assert not isinstance(votable.get_first_table().fields[0].unit, u.UnrecognizedUnit)


def test_resource_structure():
    # Based on issue #1223, as reported by @astro-friedel and @RayPlante
    from astropy.io.votable import tree as vot

    vtf = vot.VOTableFile()

    r1 = vot.Resource()
    vtf.resources.append(r1)
    t1 = vot.TableElement(vtf)
    t1.name = "t1"
    t2 = vot.TableElement(vtf)
    t2.name = "t2"
    r1.tables.append(t1)
    r1.tables.append(t2)

    r2 = vot.Resource()
    vtf.resources.append(r2)
    t3 = vot.TableElement(vtf)
    t3.name = "t3"
    t4 = vot.TableElement(vtf)
    t4.name = "t4"
    r2.tables.append(t3)
    r2.tables.append(t4)

    r3 = vot.Resource()
    vtf.resources.append(r3)
    t5 = vot.TableElement(vtf)
    t5.name = "t5"
    t6 = vot.TableElement(vtf)
    t6.name = "t6"
    r3.tables.append(t5)
    r3.tables.append(t6)

    buff = io.BytesIO()
    vtf.to_xml(buff)

    buff.seek(0)
    vtf2 = parse(buff)

    assert len(vtf2.resources) == 3

    for r in range(len(vtf2.resources)):
        res = vtf2.resources[r]
        assert len(res.tables) == 2
        assert len(res.resources) == 0


def test_no_resource_check():
    output = io.StringIO()

    # We can't test xmllint, because we can't rely on it being on the
    # user's machine.
    result = validate(
        get_pkg_data_filename("data/no_resource.xml"), output, xmllint=False
    )

    assert result is False

    output.seek(0)
    output = output.readlines()

    # Uncomment to generate new groundtruth
    # with open('no_resource.txt', 'wt', encoding='utf-8') as fd:
    #     fd.write(u''.join(output))

    with open(get_pkg_data_filename("data/no_resource.txt"), encoding="utf-8") as fd:
        truth = fd.readlines()

    truth = truth[1:]
    output = output[1:-1]

    sys.stdout.writelines(
        difflib.unified_diff(truth, output, fromfile="truth", tofile="output")
    )

    assert truth == output


def test_instantiate_vowarning():
    # This used to raise a deprecation exception.
    # See https://github.com/astropy/astroquery/pull/276
    VOWarning(())


def test_custom_datatype():
    votable = parse(
        get_pkg_data_filename("data/custom_datatype.xml"),
        datatype_mapping={"bar": "int"},
    )

    table = votable.get_first_table()
    assert table.array.dtype["foo"] == np.int32


def _timesys_tests(votable):
    assert len(list(votable.iter_timesys())) == 4

    timesys = votable.get_timesys_by_id("time_frame")
    assert timesys.timeorigin == 2455197.5
    assert timesys.timescale == "TCB"
    assert timesys.refposition == "BARYCENTER"

    timesys = votable.get_timesys_by_id("mjd_origin")
    assert timesys.timeorigin == "MJD-origin"
    assert timesys.timescale == "TDB"
    assert timesys.refposition == "EMBARYCENTER"

    timesys = votable.get_timesys_by_id("jd_origin")
    assert timesys.timeorigin == "JD-origin"
    assert timesys.timescale == "TT"
    assert timesys.refposition == "HELIOCENTER"

    timesys = votable.get_timesys_by_id("no_origin")
    assert timesys.timeorigin is None
    assert timesys.timescale == "UTC"
    assert timesys.refposition == "TOPOCENTER"


def test_timesys():
    votable = parse(get_pkg_data_filename("data/timesys.xml"))
    _timesys_tests(votable)


def test_timesys_roundtrip():
    orig_votable = parse(get_pkg_data_filename("data/timesys.xml"))
    bio = io.BytesIO()
    orig_votable.to_xml(bio)
    bio.seek(0)
    votable = parse(bio)
    _timesys_tests(votable)


def test_timesys_errors():
    output = io.StringIO()
    validate(get_pkg_data_filename("data/timesys_errors.xml"), output, xmllint=False)
    outstr = output.getvalue()
    assert "E23: Invalid timeorigin attribute 'bad-origin'" in outstr
    assert "E22: ID attribute is required for all TIMESYS elements" in outstr
    assert "W48: Unknown attribute 'refposition_mispelled' on TIMESYS" in outstr


def test_get_infos_by_name():
    vot = parse(
        io.BytesIO(
            b"""
        <VOTABLE xmlns="http://www.ivoa.net/xml/VOTable/v1.3"
          xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" version="1.4">
          <RESOURCE type="results">
            <INFO name="creator-name" value="Cannon, A."/>
            <INFO name="creator-name" value="Fleming, W."/>
          </RESOURCE>
        </VOTABLE>"""
        )
    )
    infos = vot.get_infos_by_name("creator-name")
    assert [i.value for i in infos] == ["Cannon, A.", "Fleming, W."]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest

from astropy import convolution as conv
from astropy.tests.helper import check_pickling_recovery, pickle_protocol  # noqa: F401


@pytest.mark.parametrize(
    ("name", "args", "kwargs", "xfail"),
    [
        (conv.CustomKernel, [], {"array": np.random.rand(15)}, False),
        (conv.Gaussian1DKernel, [1.0], {"x_size": 5}, True),
        (conv.Gaussian2DKernel, [1.0], {"x_size": 5, "y_size": 5}, True),
    ],
)
def test_simple_object(pickle_protocol, name, args, kwargs, xfail):  # noqa: F811
    # Tests easily instantiated objects
    if xfail:
        pytest.xfail()
    original = name(*args, **kwargs)
    check_pickling_recovery(original, pickle_protocol)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from __future__ import annotations

from collections.abc import Callable
from typing import Any

import astropy.units as u

__all__ = []

FValidateCallable = Callable[[object, object, Any], Any]
_REGISTRY_FVALIDATORS: dict[str, FValidateCallable] = {}


def _register_validator(key, fvalidate=None):
    """Decorator to register a new kind of validator function.

    Parameters
    ----------
    key : str
    fvalidate : callable[[object, object, Any], Any] or None, optional
        Value validation function.

    Returns
    -------
    ``validator`` or callable[``validator``]
        if validator is None returns a function that takes and registers a
        validator. This allows ``register_validator`` to be used as a
        decorator.
    """
    if key in _REGISTRY_FVALIDATORS:
        raise KeyError(f"validator {key!r} already registered with Parameter.")

    # fvalidate directly passed
    if fvalidate is not None:
        _REGISTRY_FVALIDATORS[key] = fvalidate
        return fvalidate

    # for use as a decorator
    def register(fvalidate):
        """Register validator function.

        Parameters
        ----------
        fvalidate : callable[[object, object, Any], Any]
            Validation function.

        Returns
        -------
        ``validator``
        """
        _REGISTRY_FVALIDATORS[key] = fvalidate
        return fvalidate

    return register


# ======================================================================


@_register_validator("default")
def _validate_with_unit(cosmology, param, value):
    """Default Parameter value validator.

    Adds/converts units if Parameter has a unit.
    """
    if param.unit is not None:
        with u.add_enabled_equivalencies(param.equivalencies):
            value = u.Quantity(value, param.unit)
    return value


@_register_validator("float")
def _validate_to_float(cosmology, param, value):
    """Parameter value validator with units, and converted to float."""
    value = _validate_with_unit(cosmology, param, value)
    return float(value)


@_register_validator("scalar")
def _validate_to_scalar(cosmology, param, value):
    """"""
    value = _validate_with_unit(cosmology, param, value)
    if not value.isscalar:
        raise ValueError(f"{param.name} is a non-scalar quantity")
    return value


@_register_validator("non-negative")
def _validate_non_negative(cosmology, param, value):
    """Parameter value validator where value is a positive float."""
    value = _validate_to_float(cosmology, param, value)
    if value < 0.0:
        raise ValueError(f"{param.name} cannot be negative.")
    return value



# These are no-regression tests for PR #13572

import numpy as np
import pytest
from numpy.testing import assert_allclose

import astropy.units as u
from astropy.coordinates import earth_orientation
from astropy.time import Time


@pytest.fixture
def tt_to_test():
    return Time("2022-08-25", scale="tt")


@pytest.mark.parametrize(
    "algorithm, result",
    [(2006, 23.43633313804873), (2000, 23.43634457995851), (1980, 23.436346167704045)],
)
def test_obliquity(tt_to_test, algorithm, result):
    assert_allclose(
        earth_orientation.obliquity(tt_to_test.jd, algorithm=algorithm),
        result,
        rtol=1e-13,
    )


def test_precession_matrix_Capitaine(tt_to_test):
    assert_allclose(
        earth_orientation.precession_matrix_Capitaine(
            tt_to_test, tt_to_test + 12.345 * u.yr
        ),
        np.array(
            [
                [9.99995470e-01, -2.76086535e-03, -1.19936388e-03],
                [2.76086537e-03, +9.99996189e-01, -1.64025847e-06],
                [1.19936384e-03, -1.67103117e-06, +9.99999281e-01],
            ]
        ),
        rtol=1e-6,
    )


def test_nutation_components2000B(tt_to_test):
    assert_allclose(
        earth_orientation.nutation_components2000B(tt_to_test.jd),
        (0.4090413775522035, -5.4418953539440996e-05, 3.176996651841667e-05),
        rtol=1e-13,
    )


def test_nutation_matrix(tt_to_test):
    assert_allclose(
        earth_orientation.nutation_matrix(tt_to_test),
        np.array(
            [
                [+9.99999999e-01, +4.99295268e-05, +2.16440489e-05],
                [-4.99288392e-05, +9.99999998e-01, -3.17705068e-05],
                [-2.16456351e-05, +3.17694261e-05, +9.99999999e-01],
            ]
        ),
        rtol=1e-6,
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import sys

from numpy import get_include as get_numpy_include
from setuptools import Extension

C_CONVOLVE_PKGDIR = os.path.relpath(os.path.dirname(__file__))

extra_compile_args = ["-UNDEBUG"]
if not sys.platform.startswith("win"):
    extra_compile_args.append("-fPIC")


def get_extensions():
    # Add '-Rpass-missed=.*' to ``extra_compile_args`` when compiling with clang
    # to report missed optimizations
    sources = [
        os.path.join(C_CONVOLVE_PKGDIR, "_convolve.pyx"),
        os.path.join(C_CONVOLVE_PKGDIR, "src", "convolve.c"),
    ]
    _convolve_ext = Extension(
        name="astropy.convolution._convolve",
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
        extra_compile_args=extra_compile_args,
        include_dirs=[get_numpy_include()],
        sources=sources,
    )
    return [_convolve_ext]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from astropy import units as u
from astropy.table import QTable, hstack, join, vstack
from astropy.utils.compat.optional_deps import HAS_H5PY
from astropy.utils.masked import Masked

from .test_masked import assert_masked_equal

FILE_FORMATS = ["ecsv", "fits"]
if HAS_H5PY:
    FILE_FORMATS.append("h5")


class MaskedArrayTableSetup:
    @classmethod
    def setup_arrays(self):
        self.a = np.array([3.0, 5.0, 0.0])
        self.mask_a = np.array([True, False, False])

    @classmethod
    def setup_class(self):
        self.setup_arrays()
        self.ma = Masked(self.a, mask=self.mask_a)
        self.ma.info.format = ".1f"
        self.t = QTable([self.ma], names=["ma"])


class MaskedQuantityTableSetup(MaskedArrayTableSetup):
    @classmethod
    def setup_arrays(self):
        self.a = np.array([3.0, 5.0, 0.0]) << u.m
        self.mask_a = np.array([True, False, False])


class TestMaskedArrayTable(MaskedArrayTableSetup):
    def test_table_initialization(self):
        assert_array_equal(self.t["ma"].unmasked, self.a)
        assert_array_equal(self.t["ma"].mask, self.mask_a)
        assert repr(self.t).splitlines()[-3:] == [
            "    ———",
            "    5.0",
            "    0.0",
        ]

    def test_info_basics(self):
        assert self.t["ma"].info.name == "ma"
        assert "serialize_method" in self.t["ma"].info.attr_names
        t2 = self.t.copy()
        t2["ma"].info.format = ".2f"
        t2["ma"].info.serialize_method["fits"] = "nonsense"
        assert repr(t2).splitlines()[-3:] == [
            "    ———",
            "   5.00",
            "   0.00",
        ]
        # Check that if we slice, things get copied over correctly.
        t3 = t2[:2]
        assert t3["ma"].info.name == "ma"
        assert t3["ma"].info.format == ".2f"
        assert "serialize_method" in t3["ma"].info.attr_names
        assert t3["ma"].info.serialize_method["fits"] == "nonsense"

    @pytest.mark.parametrize("file_format", FILE_FORMATS)
    def test_table_write(self, file_format, tmp_path):
        name = tmp_path / f"a.{file_format}"
        kwargs = {}
        if file_format == "h5":
            kwargs["path"] = "trial"
            kwargs["serialize_meta"] = True

        self.t.write(name, **kwargs)
        t2 = QTable.read(name)
        assert isinstance(t2["ma"], self.ma.__class__)
        assert np.all(t2["ma"] == self.ma)
        assert np.all(t2["ma"].mask == self.mask_a)
        if file_format == "fits":
            # Imperfect roundtrip through FITS native format description.
            assert self.t["ma"].info.format in t2["ma"].info.format
        else:
            assert t2["ma"].info.format == self.t["ma"].info.format

    @pytest.mark.parametrize("serialize_method", ["data_mask", "null_value"])
    def test_table_write_serialization(self, serialize_method, tmp_path):
        name = tmp_path / "test.ecsv"
        self.t.write(name, serialize_method=serialize_method)
        with open(name) as fh:
            lines = fh.readlines()

        t2 = QTable.read(name)
        assert isinstance(t2["ma"], self.ma.__class__)

        if serialize_method == "data_mask":
            # Will data_mask, we have data and mask columns and should
            # exactly round-trip.
            assert len(lines[-1].split()) == 2
            assert_masked_equal(t2["ma"], self.ma)
        else:
            # With null_value we have just a data column with null values
            # marked, so we lost information on the data below the mask.
            assert len(lines[-1].split()) == 1
            assert np.all(t2["ma"] == self.ma)
            assert np.all(t2["ma"].mask == self.mask_a)

    def test_non_existing_serialize_method(self, tmp_path):
        name = tmp_path / "bad.ecsv"
        with pytest.raises(ValueError, match="serialize method must be"):
            self.t.write(name, serialize_method="bad_serialize_method")


class TestMaskedQuantityTable(TestMaskedArrayTable, MaskedQuantityTableSetup):
    # Runs tests from TestMaskedArrayTable as well as some extra ones.
    def test_table_operations_requiring_masking(self):
        t1 = self.t
        t2 = QTable({"ma2": Masked([1, 2] * u.m)})
        t12 = hstack([t1, t2], join_type="outer")
        assert np.all(t12["ma"].mask == [True, False, False])
        # 'ma2' is shorter by one so we expect one True from hstack so length matches
        assert np.all(t12["ma2"].mask == [False, False, True])

        t12 = hstack([t1, t2], join_type="inner")
        assert np.all(t12["ma"].mask == [True, False])
        assert np.all(t12["ma2"].mask == [False, False])

        # Vstack tables with different column names. In this case we get masked
        # values
        t12 = vstack([t1, t2], join_type="outer")
        #  ma ma2
        #  m   m
        # --- ---
        #  ——  ——
        # 5.0  ——
        # 0.0  ——
        #  —— 1.0
        #  —— 2.0
        assert np.all(t12["ma"].mask == [True, False, False, True, True])
        assert np.all(t12["ma2"].mask == [True, True, True, False, False])

    def test_table_operations_requiring_masking_auto_promote(self):
        MaskedQuantity = Masked(u.Quantity)
        t1 = QTable({"ma1": [1, 2] * u.m})
        t2 = QTable({"ma2": [3, 4, 5] * u.m})
        t12 = hstack([t1, t2], join_type="outer")
        assert isinstance(t12["ma1"], MaskedQuantity)
        assert np.all(t12["ma1"].mask == [False, False, True])
        assert np.all(t12["ma1"] == [1, 2, 0] * u.m)
        assert not isinstance(t12["ma2"], MaskedQuantity)
        assert isinstance(t12["ma2"], u.Quantity)
        assert np.all(t12["ma2"] == [3, 4, 5] * u.m)

        t12 = hstack([t1, t2], join_type="inner")
        assert isinstance(t12["ma1"], u.Quantity)
        assert not isinstance(t12["ma1"], MaskedQuantity)
        assert isinstance(t12["ma2"], u.Quantity)
        assert not isinstance(t12["ma2"], MaskedQuantity)

        # Vstack tables with different column names. In this case we get masked
        # values
        t12 = vstack([t1, t2], join_type="outer")
        assert np.all(t12["ma1"].mask == [False, False, True, True, True])
        assert np.all(t12["ma2"].mask == [True, True, False, False, False])

        t1["a"] = [1, 2]
        t2["a"] = [1, 3, 4]
        t12 = join(t1, t2, join_type="outer")
        assert np.all(t12["ma1"].mask == [False, False, True, True])
        assert np.all(t12["ma2"].mask == [False, True, False, False])



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Handles the "Unicode" unit format.
"""

from . import console, utils


class Unicode(console.Console):
    """
    Output-only format to display pretty formatting at the console
    using Unicode characters.

    For example::

      >>> import astropy.units as u
      >>> print(u.bar.decompose().to_string('unicode'))
      100000 kg m⁻¹ s⁻²
      >>> print(u.bar.decompose().to_string('unicode', fraction='multiline'))
              kg
      100000 ────
             m s²
      >>> print(u.bar.decompose().to_string('unicode', fraction='inline'))
      100000 kg / (m s²)
    """

    _times = "×"
    _line = "─"

    @classmethod
    def _format_mantissa(cls, m):
        return m.replace("-", "−")

    @classmethod
    def _format_unit_power(cls, unit, power=1):
        name = cls._get_unit_name(unit)
        # Check for superscript units
        if power != 1:
            if name in ("°", "e⁻", "″", "′", "ʰ"):
                name = unit.short_names[0]
            name += cls._format_superscript(utils.format_power(power))
        return name

    @classmethod
    def _format_superscript(cls, number):
        mapping = str.maketrans(
            {
                "0": "⁰",
                "1": "¹",
                "2": "²",
                "3": "³",
                "4": "⁴",
                "5": "⁵",
                "6": "⁶",
                "7": "⁷",
                "8": "⁸",
                "9": "⁹",
                "-": "⁻",
                "−": "⁻",
                # This is actually a "raised omission bracket", but it's
                # the closest thing I could find to a superscript solidus.
                "/": "⸍",
            }
        )
        return number.translate(mapping)



import abc

import numpy as np

from astropy.timeseries.binned import BinnedTimeSeries
from astropy.timeseries.sampled import TimeSeries

__all__ = ["BasePeriodogram"]


class BasePeriodogram:
    @abc.abstractmethod
    def __init__(self, t, y, dy=None):
        pass

    @classmethod
    def from_timeseries(
        cls, timeseries, signal_column_name=None, uncertainty=None, **kwargs
    ):
        """
        Initialize a periodogram from a time series object.

        If a binned time series is passed, the time at the center of the bins is
        used. Also note that this method automatically gets rid of NaN/undefined
        values when initializing the periodogram.

        Parameters
        ----------
        signal_column_name : str
            The name of the column containing the signal values to use.
        uncertainty : str or float or `~astropy.units.Quantity`, optional
            The name of the column containing the errors on the signal, or the
            value to use for the error, if a scalar.
        **kwargs
            Additional keyword arguments are passed to the initializer for this
            periodogram class.
        """
        if signal_column_name is None:
            raise ValueError("signal_column_name should be set to a valid column name")

        y = timeseries[signal_column_name]
        keep = ~np.isnan(y)

        if isinstance(uncertainty, str):
            dy = timeseries[uncertainty]
            keep &= ~np.isnan(dy)
            dy = dy[keep]
        else:
            dy = uncertainty

        if isinstance(timeseries, TimeSeries):
            time = timeseries.time
        elif isinstance(timeseries, BinnedTimeSeries):
            time = timeseries.time_bin_center
        else:
            raise TypeError(
                "Input time series should be an instance of "
                "TimeSeries or BinnedTimeSeries"
            )

        return cls(time[keep], y[keep], dy=dy, **kwargs)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Functions for `astropy.cosmology`."""

from .comparison import cosmology_equal
from .optimize import z_at_value

__all__ = ["z_at_value", "cosmology_equal"]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Tests for spline models and fitters"""

import unittest.mock as mk

import numpy as np
import pytest
from numpy.testing import assert_allclose

from astropy.modeling.core import FittableModel, ModelDefinitionError
from astropy.modeling.fitting import (
    SplineExactKnotsFitter,
    SplineInterpolateFitter,
    SplineSmoothingFitter,
    SplineSplrepFitter,
)
from astropy.modeling.parameters import Parameter
from astropy.modeling.spline import Spline1D, _Spline, _SplineFitter
from astropy.utils.compat.optional_deps import HAS_SCIPY
from astropy.utils.exceptions import AstropyUserWarning

npts = 50
nknots = 10
np.random.seed(42)
test_w = np.random.rand(npts)
test_t = [-1, 0, 1]
noise = np.random.randn(npts)

degree_tests = [1, 2, 3, 4, 5]
wieght_tests = [None, test_w]
smoothing_tests = [None, 0.01]


class TestSpline:
    def setup_class(self):
        self.num_opt = 3
        self.optional_inputs = {f"test{i}": mk.MagicMock() for i in range(self.num_opt)}
        self.extra_kwargs = {f"new{i}": mk.MagicMock() for i in range(self.num_opt)}

        class Spline(_Spline):
            optional_inputs = {"test": "test"}

            def _init_parameters(self):
                super()._init_parameters()

            def _init_data(self, knots, coeffs, bounds=None):
                super()._init_data(knots, coeffs, bounds=bounds)

        self.Spline = Spline

    def test___init__(self):
        # empty spline
        spl = self.Spline()
        assert spl._t is None
        assert spl._c is None
        assert spl._user_knots is False
        assert spl._degree is None
        assert spl._test is None

        assert not hasattr(spl, "degree")

        # Call _init_spline
        with mk.patch.object(_Spline, "_init_spline", autospec=True) as mkInit:
            # No call (knots=None)
            spl = self.Spline()
            assert mkInit.call_args_list == []

            knots = mk.MagicMock()
            coeffs = mk.MagicMock()
            bounds = mk.MagicMock()
            spl = self.Spline(knots=knots, coeffs=coeffs, bounds=bounds)
            assert mkInit.call_args_list == [mk.call(spl, knots, coeffs, bounds)]

            assert spl._t is None
            assert spl._c is None
            assert spl._user_knots is False
            assert spl._degree is None
            assert spl._test is None

        # Coeffs but no knots
        MESSAGE = r"If one passes a coeffs vector one needs to also pass knots!"
        with pytest.raises(ValueError, match=MESSAGE):
            self.Spline(coeffs=mk.MagicMock())

    def test_param_names(self):
        # no parameters
        spl = self.Spline()
        assert spl.param_names == ()

        knot_names = tuple(mk.MagicMock() for _ in range(3))
        spl._knot_names = knot_names
        assert spl.param_names == knot_names

        coeff_names = tuple(mk.MagicMock() for _ in range(3))
        spl._coeff_names = coeff_names
        assert spl.param_names == knot_names + coeff_names

    def test__optional_arg(self):
        spl = self.Spline()
        assert spl._optional_arg("test") == "_test"

    def test__create_optional_inputs(self):
        class Spline(self.Spline):
            optional_inputs = self.optional_inputs

            def __init__(self):
                self._create_optional_inputs()

        spl = Spline()
        for arg in self.optional_inputs:
            attribute = spl._optional_arg(arg)
            assert hasattr(spl, attribute)
            assert getattr(spl, attribute) is None

        with pytest.raises(
            ValueError, match=r"Optional argument .* already exists in this class!"
        ):
            spl._create_optional_inputs()

    def test__intercept_optional_inputs(self):
        class Spline(self.Spline):
            optional_inputs = self.optional_inputs

            def __init__(self):
                self._create_optional_inputs()

        spl = Spline()
        new_kwargs = spl._intercept_optional_inputs(**self.extra_kwargs)
        for arg, value in self.optional_inputs.items():
            attribute = spl._optional_arg(arg)
            assert getattr(spl, attribute) is None
        assert new_kwargs == self.extra_kwargs

        kwargs = self.extra_kwargs.copy()
        for arg in self.optional_inputs:
            kwargs[arg] = mk.MagicMock()
        new_kwargs = spl._intercept_optional_inputs(**kwargs)
        for arg, value in self.optional_inputs.items():
            attribute = spl._optional_arg(arg)
            assert getattr(spl, attribute) is not None
            assert getattr(spl, attribute) == kwargs[arg]
            assert getattr(spl, attribute) != value
            assert arg not in new_kwargs
        assert new_kwargs == self.extra_kwargs
        assert kwargs != self.extra_kwargs

        with pytest.raises(
            RuntimeError, match=r".* has already been set, something has gone wrong!"
        ):
            spl._intercept_optional_inputs(**kwargs)

    def test_evaluate(self):
        class Spline(self.Spline):
            optional_inputs = self.optional_inputs

        spl = Spline()

        # No options passed in and No options set
        new_kwargs = spl.evaluate(**self.extra_kwargs)
        for arg, value in self.optional_inputs.items():
            assert new_kwargs[arg] == value
        for arg, value in self.extra_kwargs.items():
            assert new_kwargs[arg] == value
        assert len(new_kwargs) == (len(self.optional_inputs) + len(self.extra_kwargs))

        # No options passed in and Options set
        kwargs = self.extra_kwargs.copy()
        for arg in self.optional_inputs:
            kwargs[arg] = mk.MagicMock()
        spl._intercept_optional_inputs(**kwargs)
        new_kwargs = spl.evaluate(**self.extra_kwargs)
        assert new_kwargs == kwargs
        for arg in self.optional_inputs:
            attribute = spl._optional_arg(arg)
            assert getattr(spl, attribute) is None

        # Options passed in
        set_kwargs = self.extra_kwargs.copy()
        for arg in self.optional_inputs:
            kwargs[arg] = mk.MagicMock()
        spl._intercept_optional_inputs(**set_kwargs)
        kwargs = self.extra_kwargs.copy()
        for arg in self.optional_inputs:
            kwargs[arg] = mk.MagicMock()
        assert set_kwargs != kwargs
        new_kwargs = spl.evaluate(**kwargs)
        assert new_kwargs == kwargs

    def test___call__(self):
        spl = self.Spline()

        args = tuple(mk.MagicMock() for _ in range(3))
        kwargs = {f"test{idx}": mk.MagicMock() for idx in range(3)}
        new_kwargs = {f"new_test{idx}": mk.MagicMock() for idx in range(3)}
        with mk.patch.object(
            _Spline,
            "_intercept_optional_inputs",
            autospec=True,
            return_value=new_kwargs,
        ) as mkIntercept:
            with mk.patch.object(FittableModel, "__call__", autospec=True) as mkCall:
                assert mkCall.return_value == spl(*args, **kwargs)
                assert mkCall.call_args_list == [mk.call(spl, *args, **new_kwargs)]
                assert mkIntercept.call_args_list == [mk.call(spl, **kwargs)]

    def test__create_parameter(self):
        np.random.seed(37)
        base_vec = np.random.random(20)
        test = base_vec.copy()
        fixed_test = base_vec.copy()

        class Spline(self.Spline):
            @property
            def test(self):
                return test

            @property
            def fixed_test(self):
                return fixed_test

        spl = Spline()
        assert (spl.test == test).all()
        assert (spl.fixed_test == fixed_test).all()

        for index in range(20):
            name = f"test_name{index}"
            spl._create_parameter(name, index, "test")
            assert hasattr(spl, name)
            param = getattr(spl, name)
            assert isinstance(param, Parameter)
            assert param.model == spl
            assert param.fixed is False
            assert param.value == test[index] == spl.test[index] == base_vec[index]
            new_set = np.random.random()
            param.value = new_set
            assert spl.test[index] == new_set
            assert spl.test[index] != base_vec[index]
            new_get = np.random.random()
            spl.test[index] = new_get
            assert param.value == new_get
            assert param.value != new_set

        for index in range(20):
            name = f"fixed_test_name{index}"
            spl._create_parameter(name, index, "fixed_test", True)
            assert hasattr(spl, name)
            param = getattr(spl, name)
            assert isinstance(param, Parameter)
            assert param.model == spl
            assert param.fixed is True
            assert (
                param.value
                == fixed_test[index]
                == spl.fixed_test[index]
                == base_vec[index]
            )
            new_set = np.random.random()
            param.value = new_set
            assert spl.fixed_test[index] == new_set
            assert spl.fixed_test[index] != base_vec[index]
            new_get = np.random.random()
            spl.fixed_test[index] = new_get
            assert param.value == new_get
            assert param.value != new_set

    def test__create_parameters(self):
        np.random.seed(37)
        test = np.random.random(20)

        class Spline(self.Spline):
            @property
            def test(self):
                return test

        spl = Spline()

        fixed = mk.MagicMock()
        with mk.patch.object(_Spline, "_create_parameter", autospec=True) as mkCreate:
            params = spl._create_parameters("test_param", "test", fixed)
            assert params == tuple(f"test_param{idx}" for idx in range(20))
            assert mkCreate.call_args_list == [
                mk.call(spl, f"test_param{idx}", idx, "test", fixed)
                for idx in range(20)
            ]

    def test__init_parameters(self):
        spl = self.Spline()

        MESSAGE = r"This needs to be implemented"
        with pytest.raises(NotImplementedError, match=MESSAGE):
            spl._init_parameters()

    def test__init_data(self):
        spl = self.Spline()

        MESSAGE = r"This needs to be implemented"
        with pytest.raises(NotImplementedError, match=MESSAGE):
            spl._init_data(mk.MagicMock(), mk.MagicMock(), mk.MagicMock())

        with pytest.raises(NotImplementedError, match=MESSAGE):
            spl._init_data(mk.MagicMock(), mk.MagicMock())

    def test__init_spline(self):
        spl = self.Spline()

        knots = mk.MagicMock()
        coeffs = mk.MagicMock()
        bounds = mk.MagicMock()
        with mk.patch.object(
            _Spline, "_init_parameters", autospec=True
        ) as mkParameters:
            with mk.patch.object(_Spline, "_init_data", autospec=True) as mkData:
                main = mk.MagicMock()
                main.attach_mock(mkParameters, "parameters")
                main.attach_mock(mkData, "data")

                spl._init_spline(knots, coeffs, bounds)
                assert main.mock_calls == [
                    mk.call.data(spl, knots, coeffs, bounds=bounds),
                    mk.call.parameters(spl),
                ]

    def test__init_tck(self):
        spl = self.Spline()
        assert spl._c is None
        assert spl._t is None
        assert spl._degree is None

        spl = self.Spline(degree=4)
        assert spl._c is None
        assert spl._t is None
        assert spl._degree == 4


@pytest.mark.skipif(not HAS_SCIPY, reason="requires scipy")
class TestSpline1D:
    def setup_class(self):
        def func(x, noise=0):
            return np.exp(-(x**2)) + 0.1 * noise

        self.x = np.linspace(-3, 3, npts)
        self.y = func(self.x, noise)
        self.truth = func(self.x)

        arg_sort = np.argsort(self.x)
        np.random.shuffle(arg_sort)

        self.x_s = self.x[arg_sort]
        self.y_s = func(self.x_s, noise[arg_sort])

        self.npts_out = 1000
        self.xs = np.linspace(-3, 3, self.npts_out)

        self.t = np.linspace(-3, 3, nknots)[1:-1]

    def check_parameter(self, spl, base_name, name, index, value, fixed):
        assert base_name in name
        assert index == int(name.split(base_name)[-1])
        knot_name = f"{base_name}{index}"
        assert knot_name == name
        assert hasattr(spl, name)
        param = getattr(spl, name)
        assert isinstance(param, Parameter)
        assert param.name == name
        assert param.value == value(index)
        assert param.model == spl
        assert param.fixed is fixed

    def check_parameters(self, spl, params, base_name, value, fixed):
        for idx, name in enumerate(params):
            self.check_parameter(spl, base_name, name, idx, value, fixed)

    def update_parameters(self, spl, knots, value):
        for name in knots:
            param = getattr(spl, name)
            param.value = value
            assert param.value == value

    def test___init__with_no_knot_information(self):
        spl = Spline1D()
        assert spl._degree == 3
        assert spl._user_knots is False
        assert spl._t is None
        assert spl._c is None
        assert spl._nu is None

        # Check no parameters created
        assert len(spl._knot_names) == 0
        assert len(spl._coeff_names) == 0

    def test___init__with_number_of_knots(self):
        spl = Spline1D(knots=10)

        # Check baseline data
        assert spl._degree == 3
        assert spl._user_knots is False
        assert spl._nu is None

        # Check vector data
        assert len(spl._t) == 18
        t = np.zeros(18)
        t[-4:] = 1
        assert (spl._t == t).all()
        assert len(spl._c) == 18
        assert (spl._c == np.zeros(18)).all()

        # Check all parameter names created:
        assert len(spl._knot_names) == 18
        assert len(spl._coeff_names) == 18

        # Check knot values:
        def value0(idx):
            if idx < 18 - 4:
                return 0
            else:
                return 1

        self.check_parameters(spl, spl._knot_names, "knot", value0, True)

        # Check coeff values:
        def value1(idx):
            return 0

        self.check_parameters(spl, spl._coeff_names, "coeff", value1, False)

    def test___init__with_full_custom_knots(self):
        t = 17 * np.arange(20) - 32
        spl = Spline1D(knots=t)

        # Check baseline data
        assert spl._degree == 3
        assert spl._user_knots is True
        assert spl._nu is None

        # Check vector data
        assert (spl._t == t).all()
        assert len(spl._c) == 20
        assert (spl._c == np.zeros(20)).all()

        # Check all parameter names created
        assert len(spl._knot_names) == 20
        assert len(spl._coeff_names) == 20

        # Check knot values:
        def value0(idx):
            return t[idx]

        self.check_parameters(spl, spl._knot_names, "knot", value0, True)

        # Check coeff values
        def value1(idx):
            return 0

        self.check_parameters(spl, spl._coeff_names, "coeff", value1, False)

    def test___init__with_interior_custom_knots(self):
        t = np.arange(1, 20)
        spl = Spline1D(knots=t, bounds=[0, 20])
        # Check baseline data
        assert spl._degree == 3
        assert spl._user_knots is True
        assert spl._nu is None

        # Check vector data
        assert len(spl._t) == 27
        assert (spl._t[4:-4] == t).all()
        assert (spl._t[:4] == 0).all()
        assert (spl._t[-4:] == 20).all()

        assert len(spl._c) == 27
        assert (spl._c == np.zeros(27)).all()

        # Check knot values:
        def value0(idx):
            if idx < 4:
                return 0
            elif idx >= 19 + 4:
                return 20
            else:
                return t[idx - 4]

        self.check_parameters(spl, spl._knot_names, "knot", value0, True)

        # Check coeff values
        def value1(idx):
            return 0

        self.check_parameters(spl, spl._coeff_names, "coeff", value1, False)

    def test___init__with_user_knots_and_coefficients(self):
        t = 17 * np.arange(20) - 32
        c = np.linspace(-1, 1, 20)
        spl = Spline1D(knots=t, coeffs=c)

        # Check baseline data
        assert spl._degree == 3
        assert spl._user_knots is True
        assert spl._nu is None

        # Check vector data
        assert (spl._t == t).all()
        assert len(spl._c) == 20
        assert (spl._c == c).all()

        # Check all parameter names created
        assert len(spl._knot_names) == 20
        assert len(spl._coeff_names) == 20

        # Check knot values:
        def value0(idx):
            return t[idx]

        self.check_parameters(spl, spl._knot_names, "knot", value0, True)

        # Check coeff values
        def value1(idx):
            return c[idx]

        self.check_parameters(spl, spl._coeff_names, "coeff", value1, False)

    def test___init__errors(self):
        # Bad knot type
        knots = 3.5
        MESSAGE = f"Knots: {knots} must be iterable or value"
        with pytest.raises(ValueError, match=MESSAGE):
            Spline1D(knots=knots)

        # Not enough knots
        MESSAGE = r"Must have at least 8 knots"
        for idx in range(8):
            with pytest.raises(ValueError, match=MESSAGE):
                Spline1D(knots=np.arange(idx))

        # Bad scipy spline
        t = np.arange(20)[::-1]
        MESSAGE = r"Knots must be in a non-decreasing order"
        with pytest.raises(ValueError, match=MESSAGE):
            Spline1D(knots=t)

    def test_parameter_array_link(self):
        spl = Spline1D(10)

        # Check knot base values
        def value0(idx):
            if idx < 18 - 4:
                return 0
            else:
                return 1

        self.check_parameters(spl, spl._knot_names, "knot", value0, True)

        # Check knot vector -> knot parameter link
        t = np.arange(18)
        spl._t = t.copy()

        def value1(idx):
            return t[idx]

        self.check_parameters(spl, spl._knot_names, "knot", value1, True)

        # Check knot parameter -> knot vector link
        self.update_parameters(spl, spl._knot_names, 3)
        assert (spl._t[:] == 3).all()

        # Check coeff base values
        def value2(idx):
            return 0

        self.check_parameters(spl, spl._coeff_names, "coeff", value2, False)

        # Check coeff vector -> coeff parameter link
        c = 5 * np.arange(18) + 18
        spl._c = c.copy()

        def value3(idx):
            return c[idx]

        self.check_parameters(spl, spl._coeff_names, "coeff", value3, False)

        # Check coeff parameter -> coeff vector link
        self.update_parameters(spl, spl._coeff_names, 4)
        assert (spl._c[:] == 4).all()

    def test_two_splines(self):
        spl0 = Spline1D(knots=10)
        spl1 = Spline1D(knots=15, degree=2)

        assert spl0._degree == 3
        assert len(spl0._t) == 18
        t = np.zeros(18)
        t[-4:] = 1
        assert (spl0._t == t).all()
        assert len(spl0._c) == 18
        assert (spl0._c == np.zeros(18)).all()
        assert spl1._degree == 2
        assert len(spl1._t) == 21
        t = np.zeros(21)
        t[-3:] = 1
        assert (spl1._t == t).all()
        assert len(spl1._c) == 21
        assert (spl1._c == np.zeros(21)).all()

        # Check all knot names created
        assert len(spl0._knot_names) == 18
        assert len(spl1._knot_names) == 21

        # Check knot base values
        def value0(idx):
            if idx < 18 - 4:
                return 0
            else:
                return 1

        self.check_parameters(spl0, spl0._knot_names, "knot", value0, True)

        def value1(idx):
            if idx < 21 - 3:
                return 0
            else:
                return 1

        self.check_parameters(spl1, spl1._knot_names, "knot", value1, True)

        # Check knot vector -> knot parameter link
        t0 = 7 * np.arange(18) + 27
        t1 = 11 * np.arange(21) + 19
        spl0._t[:] = t0.copy()
        spl1._t[:] = t1.copy()

        def value2(idx):
            return t0[idx]

        self.check_parameters(spl0, spl0._knot_names, "knot", value2, True)

        def value3(idx):
            return t1[idx]

        self.check_parameters(spl1, spl1._knot_names, "knot", value3, True)

        # Check knot parameter -> knot vector link
        self.update_parameters(spl0, spl0._knot_names, 3)
        self.update_parameters(spl1, spl1._knot_names, 4)
        assert (spl0._t[:] == 3).all()
        assert (spl1._t[:] == 4).all()

        # Check all coeff names created
        assert len(spl0._coeff_names) == 18
        assert len(spl1._coeff_names) == 21

        # Check coeff base values
        def value4(idx):
            return 0

        self.check_parameters(spl0, spl0._coeff_names, "coeff", value4, False)
        self.check_parameters(spl1, spl1._coeff_names, "coeff", value4, False)

        # Check coeff vector -> coeff parameter link
        c0 = 17 * np.arange(18) + 14
        c1 = 37 * np.arange(21) + 47
        spl0._c[:] = c0.copy()
        spl1._c[:] = c1.copy()

        def value5(idx):
            return c0[idx]

        self.check_parameters(spl0, spl0._coeff_names, "coeff", value5, False)

        def value6(idx):
            return c1[idx]

        self.check_parameters(spl1, spl1._coeff_names, "coeff", value6, False)

        # Check coeff parameter -> coeff vector link
        self.update_parameters(spl0, spl0._coeff_names, 5)
        self.update_parameters(spl1, spl1._coeff_names, 6)
        assert (spl0._t[:] == 3).all()
        assert (spl1._t[:] == 4).all()
        assert (spl0._c[:] == 5).all()
        assert (spl1._c[:] == 6).all()

    def test__knot_names(self):
        # no parameters
        spl = Spline1D()
        assert spl._knot_names == ()

        # some parameters
        knot_names = [f"knot{idx}" for idx in range(18)]

        spl = Spline1D(10)
        assert spl._knot_names == tuple(knot_names)

    def test__coeff_names(self):
        # no parameters
        spl = Spline1D()
        assert spl._coeff_names == ()

        # some parameters
        coeff_names = [f"coeff{idx}" for idx in range(18)]

        spl = Spline1D(10)
        assert spl._coeff_names == tuple(coeff_names)

    def test_param_names(self):
        # no parameters
        spl = Spline1D()
        assert spl.param_names == ()

        # some parameters
        knot_names = [f"knot{idx}" for idx in range(18)]
        coeff_names = [f"coeff{idx}" for idx in range(18)]
        param_names = knot_names + coeff_names

        spl = Spline1D(10)
        assert spl.param_names == tuple(param_names)

    def test_t(self):
        # no parameters
        spl = Spline1D()
        # test get
        assert spl._t is None
        assert (spl.t == [0, 0, 0, 0, 1, 1, 1, 1]).all()
        # test set
        MESSAGE = r"The model parameters must be initialized before setting knots"
        with pytest.raises(ValueError, match=MESSAGE):
            spl.t = mk.MagicMock()

        # with parameters
        spl = Spline1D(10)
        # test get
        t = np.zeros(18)
        t[-4:] = 1
        assert (spl._t == t).all()
        assert (spl.t == t).all()
        # test set
        spl.t = np.arange(18) + 15
        assert (spl._t == (np.arange(18) + 15)).all()
        assert (spl.t == (np.arange(18) + 15)).all()
        assert (spl.t != t).all()
        # set error
        MESSAGE = r"There must be exactly as many knots as previously defined"
        for idx in range(30):
            if idx == 18:
                continue
            with pytest.raises(ValueError, match=MESSAGE):
                spl.t = np.arange(idx)

    def test_c(self):
        # no parameters
        spl = Spline1D()
        # test get
        assert spl._c is None
        assert (spl.c == [0, 0, 0, 0, 0, 0, 0, 0]).all()
        # test set
        MESSAGE = r"The model parameters must be initialized before setting coeffs"
        with pytest.raises(ValueError, match=MESSAGE):
            spl.c = mk.MagicMock()

        # with parameters
        spl = Spline1D(10)
        # test get
        assert (spl._c == np.zeros(18)).all()
        assert (spl.c == np.zeros(18)).all()
        # test set
        spl.c = np.arange(18) + 15
        assert (spl._c == (np.arange(18) + 15)).all()
        assert (spl.c == (np.arange(18) + 15)).all()
        assert (spl.c != np.zeros(18)).all()
        # set error
        MESSAGE = r"There must be exactly as many coeffs as previously defined"
        for idx in range(30):
            if idx == 18:
                continue
            with pytest.raises(ValueError, match=MESSAGE):
                spl.c = np.arange(idx)

    def test_degree(self):
        # default degree
        spl = Spline1D()
        # test get
        assert spl._degree == 3
        assert spl.degree == 3
        # test set

        # non-default degree
        spl = Spline1D(degree=2)
        # test get
        assert spl._degree == 2
        assert spl.degree == 2

    def test__initialized(self):
        # no parameters
        spl = Spline1D()
        assert spl._initialized is False

        # with parameters
        spl = Spline1D(knots=10, degree=2)
        assert spl._initialized is True

    def test_tck(self):
        # no parameters
        spl = Spline1D()
        # test get
        assert (spl.t == [0, 0, 0, 0, 1, 1, 1, 1]).all()
        assert (spl.c == [0, 0, 0, 0, 0, 0, 0, 0]).all()
        assert spl.degree == 3
        tck = spl.tck
        assert (tck[0] == spl.t).all()
        assert (tck[1] == spl.c).all()
        assert tck[2] == spl.degree
        # test set
        assert spl._t is None
        assert spl._c is None
        assert spl._knot_names == ()
        assert spl._coeff_names == ()
        t = np.array([0, 0, 0, 0, 1, 2, 3, 4, 5, 5, 5, 5])
        np.random.seed(619)
        c = np.random.random(12)
        k = 3
        spl.tck = (t, c, k)
        assert (spl._t == t).all()
        assert (spl._c == c).all()
        assert spl.degree == k

        def value0(idx):
            return t[idx]

        self.check_parameters(spl, spl._knot_names, "knot", value0, True)

        def value1(idx):
            return c[idx]

        self.check_parameters(spl, spl._coeff_names, "coeff", value1, False)

        # with parameters
        spl = Spline1D(knots=10, degree=2)
        # test get
        t = np.zeros(16)
        t[-3:] = 1
        assert (spl.t == t).all()
        assert (spl.c == np.zeros(16)).all()
        assert spl.degree == 2
        tck = spl.tck
        assert (tck[0] == spl.t).all()
        assert (tck[1] == spl.c).all()
        assert tck[2] == spl.degree
        # test set
        t = 5 * np.arange(16) + 11
        c = 7 * np.arange(16) + 13
        k = 2
        spl.tck = (t, c, k)
        assert (spl.t == t).all()
        assert (spl.c == c).all()
        assert spl.degree == k
        tck = spl.tck
        assert (tck[0] == spl.t).all()
        assert (tck[1] == spl.c).all()
        assert tck[2] == spl.degree

        # Error
        MESSAGE = r"tck has incompatible degree!"
        with pytest.raises(ValueError, match=MESSAGE):
            spl.tck = (t, c, 4)

    def test_bspline(self):
        from scipy.interpolate import BSpline

        # no parameters
        spl = Spline1D()
        bspline = spl.bspline

        assert isinstance(bspline, BSpline)
        assert (bspline.tck[0] == spl.tck[0]).all()
        assert (bspline.tck[1] == spl.tck[1]).all()
        assert bspline.tck[2] == spl.tck[2]

        t = np.array([0, 0, 0, 0, 1, 2, 3, 4, 5, 5, 5, 5])
        np.random.seed(619)
        c = np.random.random(12)
        k = 3

        def value0(idx):
            return t[idx]

        def value1(idx):
            return c[idx]

        # set (bspline)
        spl = Spline1D()
        assert spl._t is None
        assert spl._c is None
        assert spl._knot_names == ()
        assert spl._coeff_names == ()
        bspline = BSpline(t, c, k)
        spl.bspline = bspline
        assert (spl._t == t).all()
        assert (spl._c == c).all()
        assert spl.degree == k
        self.check_parameters(spl, spl._knot_names, "knot", value0, True)
        self.check_parameters(spl, spl._coeff_names, "coeff", value1, False)

        # set (tuple spline)
        spl = Spline1D()
        assert spl._t is None
        assert spl._c is None
        assert spl._knot_names == ()
        assert spl._coeff_names == ()
        spl.bspline = (t, c, k)
        assert (spl._t == t).all()
        assert (spl._c == c).all()
        assert spl.degree == k
        self.check_parameters(spl, spl._knot_names, "knot", value0, True)
        self.check_parameters(spl, spl._coeff_names, "coeff", value1, False)

        # with parameters
        spl = Spline1D(knots=10, degree=2)
        bspline = spl.bspline

        assert isinstance(bspline, BSpline)
        assert (bspline.tck[0] == spl.tck[0]).all()
        assert (bspline.tck[1] == spl.tck[1]).all()
        assert bspline.tck[2] == spl.tck[2]

    def test_knots(self):
        # no parameters
        spl = Spline1D()
        assert spl.knots == []

        # with parameters
        spl = Spline1D(10)
        knots = spl.knots
        assert len(knots) == 18

        for knot in knots:
            assert isinstance(knot, Parameter)
            assert hasattr(spl, knot.name)
            assert getattr(spl, knot.name) == knot

    def test_coeffs(self):
        # no parameters
        spl = Spline1D()
        assert spl.coeffs == []

        # with parameters
        spl = Spline1D(10)
        coeffs = spl.coeffs
        assert len(coeffs) == 18

        for coeff in coeffs:
            assert isinstance(coeff, Parameter)
            assert hasattr(spl, coeff.name)
            assert getattr(spl, coeff.name) == coeff

    def test__init_parameters(self):
        spl = Spline1D()

        with mk.patch.object(Spline1D, "_create_parameters", autospec=True) as mkCreate:
            spl._init_parameters()
            assert mkCreate.call_args_list == [
                mk.call(spl, "knot", "t", fixed=True),
                mk.call(spl, "coeff", "c"),
            ]

    def test__init_bounds(self):
        spl = Spline1D()

        has_bounds, lower, upper = spl._init_bounds()
        assert has_bounds is False
        assert (lower == [0, 0, 0, 0]).all()
        assert (upper == [1, 1, 1, 1]).all()
        assert spl._user_bounding_box is None

        has_bounds, lower, upper = spl._init_bounds((-5, 5))
        assert has_bounds is True
        assert (lower == [-5, -5, -5, -5]).all()
        assert (upper == [5, 5, 5, 5]).all()
        assert spl._user_bounding_box == (-5, 5)

    def test__init_knots(self):
        np.random.seed(19)
        lower = np.random.random(4)
        upper = np.random.random(4)

        # Integer
        with mk.patch.object(
            Spline1D, "bspline", new_callable=mk.PropertyMock
        ) as mkBspline:
            spl = Spline1D()
            assert spl._t is None
            spl._init_knots(10, mk.MagicMock(), lower, upper)
            t = np.concatenate((lower, np.zeros(10), upper))
            assert (spl._t == t).all()
            assert mkBspline.call_args_list == [mk.call()]

        # vector with bounds
        with mk.patch.object(
            Spline1D, "bspline", new_callable=mk.PropertyMock
        ) as mkBspline:
            knots = np.random.random(10)
            spl = Spline1D()
            assert spl._t is None
            spl._init_knots(knots, True, lower, upper)
            t = np.concatenate((lower, knots, upper))
            assert (spl._t == t).all()
            assert mkBspline.call_args_list == [mk.call()]

        # vector with no bounds
        with mk.patch.object(
            Spline1D, "bspline", new_callable=mk.PropertyMock
        ) as mkBspline:
            knots = np.random.random(10)
            spl = Spline1D()
            assert spl._t is None
            spl._init_knots(knots, False, lower, upper)
            assert (spl._t == knots).all()
            assert mkBspline.call_args_list == [mk.call()]

            # error
            MESSAGE = r"Must have at least 8 knots"
            for num in range(8):
                knots = np.random.random(num)
                spl = Spline1D()
                assert spl._t is None
                with pytest.raises(ValueError, match=MESSAGE):
                    spl._init_knots(knots, False, lower, upper)

        # Error
        spl = Spline1D()
        assert spl._t is None
        MESSAGE = r"Knots: 0.5 must be iterable or value"
        with pytest.raises(ValueError, match=MESSAGE):
            spl._init_knots(0.5, False, lower, upper)

    def test__init_coeffs(self):
        np.random.seed(492)
        # No coeffs
        with mk.patch.object(
            Spline1D, "bspline", new_callable=mk.PropertyMock
        ) as mkBspline:
            spl = Spline1D()
            assert spl._c is None
            spl._t = [1, 2, 3, 4]
            spl._init_coeffs()
            assert (spl._c == [0, 0, 0, 0]).all()
            assert mkBspline.call_args_list == [mk.call()]

        # Some coeffs
        with mk.patch.object(
            Spline1D, "bspline", new_callable=mk.PropertyMock
        ) as mkBspline:
            coeffs = np.random.random(10)
            spl = Spline1D()
            assert spl._c is None
            spl._init_coeffs(coeffs)
            assert (spl._c == coeffs).all()
            assert mkBspline.call_args_list == [mk.call()]

    def test__init_data(self):
        spl = Spline1D()

        knots = mk.MagicMock()
        coeffs = mk.MagicMock()
        bounds = mk.MagicMock()
        has_bounds = mk.MagicMock()
        lower = mk.MagicMock()
        upper = mk.MagicMock()
        with mk.patch.object(
            Spline1D,
            "_init_bounds",
            autospec=True,
            return_value=(has_bounds, lower, upper),
        ) as mkBounds:
            with mk.patch.object(Spline1D, "_init_knots", autospec=True) as mkKnots:
                with mk.patch.object(
                    Spline1D, "_init_coeffs", autospec=True
                ) as mkCoeffs:
                    main = mk.MagicMock()
                    main.attach_mock(mkBounds, "bounds")
                    main.attach_mock(mkKnots, "knots")
                    main.attach_mock(mkCoeffs, "coeffs")

                    spl._init_data(knots, coeffs, bounds)
                    assert main.mock_calls == [
                        mk.call.bounds(spl, bounds),
                        mk.call.knots(spl, knots, has_bounds, lower, upper),
                        mk.call.coeffs(spl, coeffs),
                    ]

    def test_evaluate(self):
        spl = Spline1D()

        args = tuple(mk.MagicMock() for _ in range(3))
        kwargs = {f"test{idx}": mk.MagicMock() for idx in range(3)}
        new_kwargs = {f"new_test{idx}": mk.MagicMock() for idx in range(3)}

        with mk.patch.object(
            _Spline, "evaluate", autospec=True, return_value=new_kwargs
        ) as mkEval:
            with mk.patch.object(
                Spline1D, "bspline", new_callable=mk.PropertyMock
            ) as mkBspline:
                assert mkBspline.return_value.return_value == spl.evaluate(
                    *args, **kwargs
                )
                assert mkBspline.return_value.call_args_list == [
                    mk.call(args[0], **new_kwargs)
                ]
                assert mkBspline.call_args_list == [mk.call()]
                assert mkEval.call_args_list == [mk.call(spl, *args, **kwargs)]

        # Error
        MESSAGE = r"Cannot evaluate a derivative of order higher than 4"
        for idx in range(5, 8):
            with mk.patch.object(
                _Spline, "evaluate", autospec=True, return_value={"nu": idx}
            ):
                with pytest.raises(RuntimeError, match=MESSAGE):
                    spl.evaluate(*args, **kwargs)

    def check_knots_created(self, spl, k):
        def value0(idx):
            return self.x[0]

        def value1(idx):
            return self.x[-1]

        for idx in range(k + 1):
            name = f"knot{idx}"
            self.check_parameter(spl, "knot", name, idx, value0, True)

            index = len(spl.t) - (k + 1) + idx
            name = f"knot{index}"
            self.check_parameter(spl, "knot", name, index, value1, True)

        def value3(idx):
            return spl.t[idx]

        assert len(spl._knot_names) == len(spl.t)
        for idx, name in enumerate(spl._knot_names):
            assert name == f"knot{idx}"
            self.check_parameter(spl, "knot", name, idx, value3, True)

    def check_coeffs_created(self, spl):
        def value(idx):
            return spl.c[idx]

        assert len(spl._coeff_names) == len(spl.c)
        for idx, name in enumerate(spl._coeff_names):
            assert name == f"coeff{idx}"
            self.check_parameter(spl, "coeff", name, idx, value, False)

    @staticmethod
    def check_base_spline(spl, t, c, k):
        """Check the base spline form"""
        if t is None:
            assert spl._t is None
        else:
            assert_allclose(spl._t, t)

        if c is None:
            assert spl._c is None
        else:
            assert_allclose(spl._c, c)

        assert spl.degree == k
        assert spl._bounding_box is None

    def check_spline_fit(self, fit_spl, spline, fitter, atol_fit, atol_truth):
        """Check the spline fit"""
        assert_allclose(fit_spl.t, spline._eval_args[0])
        assert_allclose(fit_spl.c, spline._eval_args[1])
        assert_allclose(fitter.fit_info["spline"]._eval_args[0], spline._eval_args[0])
        assert_allclose(fitter.fit_info["spline"]._eval_args[1], spline._eval_args[1])

        # check that _parameters are correct
        assert len(fit_spl._parameters) == len(fit_spl.t) + len(fit_spl.c)
        assert_allclose(fit_spl._parameters[: len(fit_spl.t)], fit_spl.t)
        assert_allclose(fit_spl._parameters[len(fit_spl.t) :], fit_spl.c)

        # check that parameters are correct
        assert len(fit_spl.parameters) == len(fit_spl.t) + len(fit_spl.c)
        assert_allclose(fit_spl.parameters[: len(fit_spl.t)], fit_spl.t)
        assert_allclose(fit_spl.parameters[len(fit_spl.t) :], fit_spl.c)

        assert_allclose(spline.get_residual(), fitter.fit_info["resid"])

        assert_allclose(fit_spl(self.x), spline(self.x))
        assert_allclose(fit_spl(self.x), fitter.fit_info["spline"](self.x))

        assert_allclose(fit_spl(self.x), self.y, atol=atol_fit)
        assert_allclose(fit_spl(self.x), self.truth, atol=atol_truth)

    def check_bbox(self, spl, fit_spl, fitter, w, **kwargs):
        """Check the spline fit with bbox option"""
        bbox = [self.x[0], self.x[-1]]
        bbox_spl = fitter(spl, self.x, self.y, weights=w, bbox=bbox, **kwargs)
        assert bbox_spl.bounding_box == tuple(bbox)
        assert_allclose(fit_spl.t, bbox_spl.t)
        assert_allclose(fit_spl.c, bbox_spl.c)

    def check_knots_warning(self, fitter, knots, k, w, **kwargs):
        """Check that the knots warning is raised"""
        spl = Spline1D(knots=knots, degree=k)
        with pytest.warns(AstropyUserWarning):
            fitter(spl, self.x, self.y, weights=w, **kwargs)

    @pytest.mark.parametrize("w", wieght_tests)
    @pytest.mark.parametrize("k", degree_tests)
    def test_interpolate_fitter(self, w, k):
        fitter = SplineInterpolateFitter()
        assert fitter.fit_info == {"resid": None, "spline": None}

        spl = Spline1D(degree=k)
        self.check_base_spline(spl, None, None, k)

        fit_spl = fitter(spl, self.x, self.y, weights=w)
        self.check_base_spline(spl, None, None, k)

        assert len(fit_spl.t) == (len(self.x) + k + 1) == len(fit_spl._knot_names)
        self.check_knots_created(fit_spl, k)
        self.check_coeffs_created(fit_spl)
        assert fit_spl._bounding_box is None

        from scipy.interpolate import InterpolatedUnivariateSpline, UnivariateSpline

        spline = InterpolatedUnivariateSpline(self.x, self.y, w=w, k=k)
        assert isinstance(fitter.fit_info["spline"], UnivariateSpline)

        assert spline.get_residual() == 0
        self.check_spline_fit(fit_spl, spline, fitter, 0, 1)
        self.check_bbox(spl, fit_spl, fitter, w)

        knots = np.linspace(self.x[0], self.x[-1], len(self.x) + k + 1)
        self.check_knots_warning(fitter, knots, k, w)

    @pytest.mark.parametrize("w", wieght_tests)
    @pytest.mark.parametrize("k", degree_tests)
    @pytest.mark.parametrize("s", smoothing_tests)
    def test_smoothing_fitter(self, w, k, s):
        fitter = SplineSmoothingFitter()
        assert fitter.fit_info == {"resid": None, "spline": None}

        spl = Spline1D(degree=k)
        self.check_base_spline(spl, None, None, k)

        fit_spl = fitter(spl, self.x, self.y, s=s, weights=w)
        self.check_base_spline(spl, None, None, k)

        self.check_knots_created(fit_spl, k)
        self.check_coeffs_created(fit_spl)
        assert fit_spl._bounding_box is None

        from scipy.interpolate import UnivariateSpline

        spline = UnivariateSpline(self.x, self.y, w=w, k=k, s=s)
        assert isinstance(fitter.fit_info["spline"], UnivariateSpline)

        self.check_spline_fit(fit_spl, spline, fitter, 1, 1)
        self.check_bbox(spl, fit_spl, fitter, w, s=s)

        # test warning
        knots = fit_spl.t.copy()
        self.check_knots_warning(fitter, knots, k, w, s=s)

    @pytest.mark.parametrize("w", wieght_tests)
    @pytest.mark.parametrize("k", degree_tests)
    def test_exact_knots_fitter(self, w, k):
        fitter = SplineExactKnotsFitter()
        assert fitter.fit_info == {"resid": None, "spline": None}

        knots = [-1, 0, 1]
        t = np.concatenate(([self.x[0]] * (k + 1), knots, [self.x[-1]] * (k + 1)))
        c = np.zeros(len(t))

        # With knots preset
        spl = Spline1D(knots=knots, degree=k, bounds=[self.x[0], self.x[-1]])
        self.check_base_spline(spl, t, c, k)
        assert (spl.t_interior == knots).all()

        fit_spl = fitter(spl, self.x, self.y, weights=w)
        self.check_base_spline(spl, t, c, k)
        assert (spl.t_interior == knots).all()

        assert len(fit_spl.t) == len(t) == len(fit_spl._knot_names)
        self.check_knots_created(fit_spl, k)
        self.check_coeffs_created(fit_spl)
        assert fit_spl._bounding_box is None

        from scipy.interpolate import LSQUnivariateSpline, UnivariateSpline

        spline = LSQUnivariateSpline(self.x, self.y, knots, w=w, k=k)
        assert isinstance(fitter.fit_info["spline"], UnivariateSpline)

        assert_allclose(spline.get_residual(), 0.1, atol=1)
        assert_allclose(fitter.fit_info["spline"].get_residual(), 0.1, atol=1)
        self.check_spline_fit(fit_spl, spline, fitter, 1, 1)
        self.check_bbox(spl, fit_spl, fitter, w)

        # Pass knots via fitter function
        with pytest.warns(AstropyUserWarning):
            fitter(spl, self.x, self.y, t=knots, weights=w)

        # pass no knots
        spl = Spline1D(degree=k)
        MESSAGE = r"No knots have been provided"
        with pytest.raises(RuntimeError, match=MESSAGE):
            fitter(spl, self.x, self.y, weights=w)

    @pytest.mark.parametrize("w", wieght_tests)
    @pytest.mark.parametrize("k", degree_tests)
    @pytest.mark.parametrize("s", smoothing_tests)
    def test_splrep_fitter_no_knots(self, w, k, s):
        fitter = SplineSplrepFitter()
        assert fitter.fit_info == {"fp": None, "ier": None, "msg": None}

        spl = Spline1D(degree=k)
        self.check_base_spline(spl, None, None, k)

        fit_spl = fitter(spl, self.x, self.y, s=s, weights=w)
        self.check_base_spline(spl, None, None, k)

        self.check_knots_created(fit_spl, k)
        self.check_coeffs_created(fit_spl)
        assert fit_spl._bounding_box is None

        from scipy.interpolate import BSpline, splrep

        tck, spline_fp, spline_ier, spline_msg = splrep(
            self.x, self.y, w=w, k=k, s=s, full_output=1
        )
        assert_allclose(fit_spl.t, tck[0])
        assert_allclose(fit_spl.c, tck[1])

        assert fitter.fit_info["fp"] == spline_fp
        assert fitter.fit_info["ier"] == spline_ier
        assert fitter.fit_info["msg"] == spline_msg

        spline = BSpline(*tck)
        assert_allclose(fit_spl(self.x), spline(self.x))

        assert_allclose(fit_spl(self.x), self.y, atol=1)
        assert_allclose(fit_spl(self.x), self.truth, atol=1)

        self.check_bbox(spl, fit_spl, fitter, w, s=s)

    @pytest.mark.parametrize("w", wieght_tests)
    @pytest.mark.parametrize("k", degree_tests)
    def test_splrep_fitter_with_knots(self, w, k):
        fitter = SplineSplrepFitter()
        assert fitter.fit_info == {"fp": None, "ier": None, "msg": None}

        knots = [-1, 0, 1]
        t = np.concatenate(([self.x[0]] * (k + 1), knots, [self.x[-1]] * (k + 1)))
        c = np.zeros(len(t))

        # With knots preset
        spl = Spline1D(knots=knots, degree=k, bounds=[self.x[0], self.x[-1]])
        self.check_base_spline(spl, t, c, k)
        assert (spl.t_interior == knots).all()

        fit_spl = fitter(spl, self.x, self.y, weights=w)
        self.check_base_spline(spl, t, c, k)
        assert (spl.t_interior == knots).all()

        self.check_knots_created(fit_spl, k)
        self.check_coeffs_created(fit_spl)
        assert fit_spl._bounding_box is None

        from scipy.interpolate import BSpline, splrep

        tck, spline_fp, spline_ier, spline_msg = splrep(
            self.x, self.y, w=w, k=k, t=knots, full_output=1
        )
        assert_allclose(fit_spl.t, tck[0])
        assert_allclose(fit_spl.c, tck[1])

        assert fitter.fit_info["fp"] == spline_fp
        assert fitter.fit_info["ier"] == spline_ier
        assert fitter.fit_info["msg"] == spline_msg

        spline = BSpline(*tck)
        assert_allclose(fit_spl(self.x), spline(self.x))

        assert_allclose(fit_spl(self.x), self.y, atol=1)
        assert_allclose(fit_spl(self.x), self.truth, atol=1)

        self.check_bbox(spl, fit_spl, fitter, w)

        # test warning
        with pytest.warns(AstropyUserWarning):
            fitter(spl, self.x, self.y, t=knots, weights=w)

        # With no knots present
        spl = Spline1D(degree=k)
        self.check_base_spline(spl, None, None, k)

        fit_spl = fitter(spl, self.x, self.y, t=knots, weights=w)
        self.check_base_spline(spl, None, None, k)

        self.check_knots_created(fit_spl, k)
        self.check_coeffs_created(fit_spl)
        assert fit_spl._bounding_box is None

        from scipy.interpolate import BSpline, splrep

        tck = splrep(self.x, self.y, w=w, k=k, t=knots)
        assert_allclose(fit_spl.t, tck[0])
        assert_allclose(fit_spl.c, tck[1])

        spline = BSpline(*tck)
        assert_allclose(fit_spl(self.x), spline(self.x))

        assert_allclose(fit_spl(self.x), self.y, atol=1)
        assert_allclose(fit_spl(self.x), self.truth, atol=1)

        self.check_bbox(spl, fit_spl, fitter, w, t=knots)

    def generate_spline(self, w=None, bbox=[None] * 2, k=None, s=None, t=None):
        if k is None:
            k = 3

        from scipy.interpolate import BSpline, splrep

        tck = splrep(self.x, self.y, w=w, xb=bbox[0], xe=bbox[1], k=k, s=s, t=t)

        return BSpline(*tck)

    def test_derivative(self):
        bspline = self.generate_spline()

        spl = Spline1D()
        spl.bspline = bspline
        assert_allclose(spl.t, bspline.t)
        assert_allclose(spl.c, bspline.c)
        assert spl.degree == bspline.k

        # 1st derivative
        d_bspline = bspline.derivative(nu=1)
        assert_allclose(d_bspline(self.xs), bspline(self.xs, nu=1))
        assert_allclose(d_bspline(self.xs, nu=1), bspline(self.xs, nu=2))
        assert_allclose(d_bspline(self.xs, nu=2), bspline(self.xs, nu=3))
        assert_allclose(d_bspline(self.xs, nu=3), bspline(self.xs, nu=4))

        der = spl.derivative()
        assert_allclose(der.t, d_bspline.t)
        assert_allclose(der.c, d_bspline.c)
        assert der.degree == d_bspline.k == 2
        assert_allclose(der.evaluate(self.xs), spl.evaluate(self.xs, nu=1))
        assert_allclose(der.evaluate(self.xs, nu=1), spl.evaluate(self.xs, nu=2))
        assert_allclose(der.evaluate(self.xs, nu=2), spl.evaluate(self.xs, nu=3))
        assert_allclose(der.evaluate(self.xs, nu=3), spl.evaluate(self.xs, nu=4))

        # 2nd derivative
        d_bspline = bspline.derivative(nu=2)
        assert_allclose(d_bspline(self.xs), bspline(self.xs, nu=2))
        assert_allclose(d_bspline(self.xs, nu=1), bspline(self.xs, nu=3))
        assert_allclose(d_bspline(self.xs, nu=2), bspline(self.xs, nu=4))

        der = spl.derivative(nu=2)
        assert_allclose(der.t, d_bspline.t)
        assert_allclose(der.c, d_bspline.c)
        assert der.degree == d_bspline.k == 1
        assert_allclose(der.evaluate(self.xs), spl.evaluate(self.xs, nu=2))
        assert_allclose(der.evaluate(self.xs, nu=1), spl.evaluate(self.xs, nu=3))
        assert_allclose(der.evaluate(self.xs, nu=2), spl.evaluate(self.xs, nu=4))

        # 3rd derivative
        d_bspline = bspline.derivative(nu=3)
        assert_allclose(d_bspline(self.xs), bspline(self.xs, nu=3))
        assert_allclose(d_bspline(self.xs, nu=1), bspline(self.xs, nu=4))

        der = spl.derivative(nu=3)
        assert_allclose(der.t, d_bspline.t)
        assert_allclose(der.c, d_bspline.c)
        assert der.degree == d_bspline.k == 0
        assert_allclose(der.evaluate(self.xs), spl.evaluate(self.xs, nu=3))
        assert_allclose(der.evaluate(self.xs, nu=1), spl.evaluate(self.xs, nu=4))

        # Too many derivatives
        MESSAGE = r"Must have nu <= 3"
        for nu in range(4, 9):
            with pytest.raises(ValueError, match=MESSAGE):
                spl.derivative(nu=nu)

    def test_antiderivative(self):
        bspline = self.generate_spline()

        spl = Spline1D()
        spl.bspline = bspline

        # 1st antiderivative
        a_bspline = bspline.antiderivative(nu=1)
        assert_allclose(bspline(self.xs), a_bspline(self.xs, nu=1))
        assert_allclose(bspline(self.xs, nu=1), a_bspline(self.xs, nu=2))
        assert_allclose(bspline(self.xs, nu=2), a_bspline(self.xs, nu=3))
        assert_allclose(bspline(self.xs, nu=3), a_bspline(self.xs, nu=4))
        assert_allclose(bspline(self.xs, nu=4), a_bspline(self.xs, nu=5))

        anti = spl.antiderivative()
        assert_allclose(anti.t, a_bspline.t)
        assert_allclose(anti.c, a_bspline.c)
        assert anti.degree == a_bspline.k == 4
        assert_allclose(spl.evaluate(self.xs), anti.evaluate(self.xs, nu=1))
        assert_allclose(spl.evaluate(self.xs, nu=1), anti.evaluate(self.xs, nu=2))
        assert_allclose(spl.evaluate(self.xs, nu=2), anti.evaluate(self.xs, nu=3))
        assert_allclose(spl.evaluate(self.xs, nu=3), anti.evaluate(self.xs, nu=4))
        assert_allclose(spl.evaluate(self.xs, nu=4), anti.evaluate(self.xs, nu=5))

        # 2nd antiderivative
        a_bspline = bspline.antiderivative(nu=2)
        assert_allclose(bspline(self.xs), a_bspline(self.xs, nu=2))
        assert_allclose(bspline(self.xs, nu=1), a_bspline(self.xs, nu=3))
        assert_allclose(bspline(self.xs, nu=2), a_bspline(self.xs, nu=4))
        assert_allclose(bspline(self.xs, nu=3), a_bspline(self.xs, nu=5))
        assert_allclose(bspline(self.xs, nu=4), a_bspline(self.xs, nu=6))

        anti = spl.antiderivative(nu=2)
        assert_allclose(anti.t, a_bspline.t)
        assert_allclose(anti.c, a_bspline.c)
        assert anti.degree == a_bspline.k == 5
        assert_allclose(spl.evaluate(self.xs), anti.evaluate(self.xs, nu=2))
        assert_allclose(spl.evaluate(self.xs, nu=1), anti.evaluate(self.xs, nu=3))
        assert_allclose(spl.evaluate(self.xs, nu=2), anti.evaluate(self.xs, nu=4))
        assert_allclose(spl.evaluate(self.xs, nu=3), anti.evaluate(self.xs, nu=5))
        assert_allclose(spl.evaluate(self.xs, nu=4), anti.evaluate(self.xs, nu=6))

        # Too many anti derivatives
        for nu in range(3, 9):
            MESSAGE = (
                "Supported splines can have max degree 5, antiderivative degree will"
                f" be {nu + 3}"
            )
            with pytest.raises(ValueError, match=MESSAGE):
                spl.antiderivative(nu=nu)

    def test__SplineFitter_error(self):
        spl = Spline1D()

        class SplineFitter(_SplineFitter):
            def _fit_method(self, model, x, y, **kwargs):
                super()._fit_method(model, x, y, **kwargs)

        fitter = SplineFitter()

        MESSAGE = r"1D model can only have 2 data points"
        with pytest.raises(ValueError, match=MESSAGE):
            fitter(spl, mk.MagicMock(), mk.MagicMock(), mk.MagicMock())

        MESSAGE = r"Only spline models are compatible with this fitter"
        with pytest.raises(ModelDefinitionError, match=MESSAGE):
            fitter(mk.MagicMock(), mk.MagicMock(), mk.MagicMock())

        MESSAGE = r"This has not been implemented for _SplineFitter"
        with pytest.raises(NotImplementedError, match=MESSAGE):
            fitter(spl, mk.MagicMock(), mk.MagicMock())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Astronomical and physics constants in SI units.  See :mod:`astropy.constants`
for a complete listing of constants defined in Astropy.
"""

import numpy as np

from .constant import Constant

# ASTRONOMICAL CONSTANTS


class IAU2012(Constant):
    default_reference = "IAU 2012"
    _registry = {}
    _has_incompatible_units = set()


# DISTANCE

# Astronomical Unit
au = IAU2012(
    "au",
    "Astronomical Unit",
    1.49597870700e11,
    "m",
    0.0,
    "IAU 2012 Resolution B2",
    system="si",
)

# Parsec

pc = IAU2012(
    "pc",
    "Parsec",
    au.value / np.tan(np.radians(1.0 / 3600.0)),
    "m",
    au.uncertainty / np.tan(np.radians(1.0 / 3600.0)),
    "Derived from au",
    system="si",
)

# Kiloparsec
kpc = IAU2012(
    "kpc",
    "Kiloparsec",
    1000.0 * au.value / np.tan(np.radians(1.0 / 3600.0)),
    "m",
    1000.0 * au.uncertainty / np.tan(np.radians(1.0 / 3600.0)),
    "Derived from au",
    system="si",
)

# Luminosity not defined till 2015 (https://arxiv.org/abs/1510.06262)
L_bol0 = IAU2012(
    "L_bol0",
    "Luminosity for absolute bolometric magnitude 0",
    3.0128e28,
    "W",
    0.0,
    "IAU 2015 Resolution B 2",
    system="si",
)


# SOLAR QUANTITIES

# Solar luminosity
L_sun = IAU2012(
    "L_sun",
    "Solar luminosity",
    3.846e26,
    "W",
    0.0005e26,
    "Allen's Astrophysical Quantities 4th Ed.",
    system="si",
)

# Solar mass
M_sun = IAU2012(
    "M_sun",
    "Solar mass",
    1.9891e30,
    "kg",
    0.00005e30,
    "Allen's Astrophysical Quantities 4th Ed.",
    system="si",
)

# Solar radius
R_sun = IAU2012(
    "R_sun",
    "Solar radius",
    6.95508e8,
    "m",
    0.00026e8,
    "Allen's Astrophysical Quantities 4th Ed.",
    system="si",
)


# OTHER SOLAR SYSTEM QUANTITIES

# Jupiter mass
M_jup = IAU2012(
    "M_jup",
    "Jupiter mass",
    1.8987e27,
    "kg",
    0.00005e27,
    "Allen's Astrophysical Quantities 4th Ed.",
    system="si",
)

# Jupiter equatorial radius
R_jup = IAU2012(
    "R_jup",
    "Jupiter equatorial radius",
    7.1492e7,
    "m",
    0.00005e7,
    "Allen's Astrophysical Quantities 4th Ed.",
    system="si",
)

# Earth mass
M_earth = IAU2012(
    "M_earth",
    "Earth mass",
    5.9742e24,
    "kg",
    0.00005e24,
    "Allen's Astrophysical Quantities 4th Ed.",
    system="si",
)

# Earth equatorial radius
R_earth = IAU2012(
    "R_earth",
    "Earth equatorial radius",
    6.378136e6,
    "m",
    0.0000005e6,
    "Allen's Astrophysical Quantities 4th Ed.",
    system="si",
)



# Licensed under a 3-clause BSD style license - see LICNSE.rst

# This module includes files automatically generated from ply (these end in
# _lextab.py and _parsetab.py). To generate these files, remove them from this
# folder, then build astropy and run the tests in-place:
#
#   python setup.py build_ext --inplace
#   pytest astropy/units
#
# You can then commit the changes to the re-generated _lextab.py and
# _parsetab.py files.

"""
Handles units in `Office of Guest Investigator Programs (OGIP)
FITS files
<https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/general/ogip_93_001/>`__.
"""

import copy
import keyword
import math
import warnings
from fractions import Fraction

from astropy.utils import parsing

from . import core, generic, utils


class OGIP(generic.Generic):
    """
    Support the units in `Office of Guest Investigator Programs (OGIP)
    FITS files
    <https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/general/ogip_93_001/>`__.
    """

    _tokens = (
        "DIVISION",
        "OPEN_PAREN",
        "CLOSE_PAREN",
        "WHITESPACE",
        "STARSTAR",
        "STAR",
        "SIGN",
        "UFLOAT",
        "LIT10",
        "UINT",
        "UNKNOWN",
        "UNIT",
    )

    @staticmethod
    def _generate_unit_names():
        from astropy import units as u

        names = {}
        deprecated_names = set()
        bases = [
            "A", "C", "cd", "eV", "F", "g", "H", "Hz", "J",
            "Jy", "K", "lm", "lx", "m", "mol", "N", "ohm", "Pa",
            "pc", "rad", "s", "S", "sr", "T", "V", "W", "Wb",
        ]  # fmt: skip
        deprecated_bases = []
        prefixes = [
            "y", "z", "a", "f", "p", "n", "u", "m", "c", "d",
            "", "da", "h", "k", "M", "G", "T", "P", "E", "Z", "Y",
        ]  # fmt: skip

        for base in bases + deprecated_bases:
            for prefix in prefixes:
                key = prefix + base
                if keyword.iskeyword(key):
                    continue
                names[key] = getattr(u, key)
        for base in deprecated_bases:
            for prefix in prefixes:
                deprecated_names.add(prefix + base)
        simple_units = [
            "angstrom", "arcmin", "arcsec", "AU", "barn", "bin",
            "byte", "chan", "count", "day", "deg", "erg", "G",
            "h", "lyr", "mag", "min", "photon", "pixel",
            "voxel", "yr",
        ]  # fmt: skip
        for unit in simple_units:
            names[unit] = getattr(u, unit)

        # Create a separate, disconnected unit for the special case of
        # Crab and mCrab, since OGIP doesn't define their quantities.
        Crab = u.def_unit(["Crab"], prefixes=False, doc="Crab (X-ray flux)")
        mCrab = u.Unit(10**-3 * Crab)
        names["Crab"] = Crab
        names["mCrab"] = mCrab

        deprecated_units = ["Crab", "mCrab"]
        for unit in deprecated_units:
            deprecated_names.add(unit)
        functions = [
            "log", "ln", "exp", "sqrt", "sin", "cos", "tan", "asin",
            "acos", "atan", "sinh", "cosh", "tanh",
        ]  # fmt: skip
        for name in functions:
            names[name] = name

        return names, deprecated_names, functions

    @classmethod
    def _make_lexer(cls):
        tokens = cls._tokens

        t_DIVISION = r"/"
        t_OPEN_PAREN = r"\("
        t_CLOSE_PAREN = r"\)"
        t_WHITESPACE = "[ \t]+"
        t_STARSTAR = r"\*\*"
        t_STAR = r"\*"

        # NOTE THE ORDERING OF THESE RULES IS IMPORTANT!!
        # Regular expression rules for simple tokens
        def t_UFLOAT(t):
            r"(((\d+\.?\d*)|(\.\d+))([eE][+-]?\d+))|(((\d+\.\d*)|(\.\d+))([eE][+-]?\d+)?)"
            t.value = float(t.value)
            return t

        def t_UINT(t):
            r"\d+"
            t.value = int(t.value)
            return t

        def t_SIGN(t):
            r"[+-](?=\d)"
            t.value = float(t.value + "1")
            return t

        def t_X(t):  # multiplication for factor in front of unit
            r"[x×]"
            return t

        def t_LIT10(t):
            r"10"
            return 10

        def t_UNKNOWN(t):
            r"[Uu][Nn][Kk][Nn][Oo][Ww][Nn]"
            return None

        def t_UNIT(t):
            r"[a-zA-Z][a-zA-Z_]*"
            t.value = cls._get_unit(t)
            return t

        # Don't ignore whitespace
        t_ignore = ""

        # Error handling rule
        def t_error(t):
            raise ValueError(f"Invalid character at col {t.lexpos}")

        return parsing.lex(lextab="ogip_lextab", package="astropy/units")

    @classmethod
    def _make_parser(cls):
        """
        The grammar here is based on the description in the
        `Specification of Physical Units within OGIP FITS files
        <https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/general/ogip_93_001/>`__,
        which is not terribly precise.  The exact grammar is here is
        based on the YACC grammar in the `unity library
        <https://bitbucket.org/nxg/unity/>`_.
        """
        tokens = cls._tokens

        def p_main(p):
            """
            main : UNKNOWN
                 | complete_expression
                 | scale_factor complete_expression
                 | scale_factor WHITESPACE complete_expression
            """
            if len(p) == 4:
                p[0] = p[1] * p[3]
            elif len(p) == 3:
                p[0] = p[1] * p[2]
            else:
                p[0] = p[1]

        def p_complete_expression(p):
            """
            complete_expression : product_of_units
            """
            p[0] = p[1]

        def p_product_of_units(p):
            """
            product_of_units : unit_expression
                             | division unit_expression
                             | product_of_units product unit_expression
                             | product_of_units division unit_expression
            """
            if len(p) == 4:
                if p[2] == "DIVISION":
                    p[0] = p[1] / p[3]
                else:
                    p[0] = p[1] * p[3]
            elif len(p) == 3:
                p[0] = p[2] ** -1
            else:
                p[0] = p[1]

        def p_unit_expression(p):
            """
            unit_expression : unit
                            | UNIT OPEN_PAREN complete_expression CLOSE_PAREN
                            | OPEN_PAREN complete_expression CLOSE_PAREN
                            | UNIT OPEN_PAREN complete_expression CLOSE_PAREN power numeric_power
                            | OPEN_PAREN complete_expression CLOSE_PAREN power numeric_power
            """
            # If we run p[1] in cls._functions, it will try and parse each
            # item in the list into a unit, which is slow. Since we know that
            # all the items in the list are strings, we can simply convert
            # p[1] to a string instead.
            p1_str = str(p[1])

            if p1_str in cls._functions and p1_str != "sqrt":
                raise ValueError(
                    f"The function '{p[1]}' is valid in OGIP, but not understood "
                    "by astropy.units."
                )

            if len(p) == 7:
                if p1_str == "sqrt":
                    p[0] = p[1] * p[3] ** (0.5 * p[6])
                else:
                    p[0] = p[1] * p[3] ** p[6]
            elif len(p) == 6:
                p[0] = p[2] ** p[5]
            elif len(p) == 5:
                if p1_str == "sqrt":
                    p[0] = p[3] ** 0.5
                else:
                    p[0] = p[1] * p[3]
            elif len(p) == 4:
                p[0] = p[2]
            else:
                p[0] = p[1]

        def p_scale_factor(p):
            """
            scale_factor : LIT10 power numeric_power
                         | LIT10
                         | signed_float
                         | signed_float power numeric_power
                         | signed_int power numeric_power
            """
            if len(p) == 4:
                p[0] = 10 ** p[3]
            else:
                p[0] = p[1]
            # Can't use np.log10 here, because p[0] may be a Python long.
            if math.log10(p[0]) % 1.0 != 0.0:
                from astropy.units.core import UnitsWarning

                warnings.warn(
                    f"'{p[0]}' scale should be a power of 10 in OGIP format",
                    UnitsWarning,
                )

        def p_division(p):
            """
            division : DIVISION
                     | WHITESPACE DIVISION
                     | WHITESPACE DIVISION WHITESPACE
                     | DIVISION WHITESPACE
            """
            p[0] = "DIVISION"

        def p_product(p):
            """
            product : WHITESPACE
                    | STAR
                    | WHITESPACE STAR
                    | WHITESPACE STAR WHITESPACE
                    | STAR WHITESPACE
            """
            p[0] = "PRODUCT"

        def p_power(p):
            """
            power : STARSTAR
            """
            p[0] = "POWER"

        def p_unit(p):
            """
            unit : UNIT
                 | UNIT power numeric_power
            """
            if len(p) == 4:
                p[0] = p[1] ** p[3]
            else:
                p[0] = p[1]

        def p_numeric_power(p):
            """
            numeric_power : UINT
                          | signed_float
                          | OPEN_PAREN signed_int CLOSE_PAREN
                          | OPEN_PAREN signed_float CLOSE_PAREN
                          | OPEN_PAREN signed_float division UINT CLOSE_PAREN
            """
            if len(p) == 6:
                p[0] = Fraction(int(p[2]), int(p[4]))
            elif len(p) == 4:
                p[0] = p[2]
            else:
                p[0] = p[1]

        def p_sign(p):
            """
            sign : SIGN
                 |
            """
            if len(p) == 2:
                p[0] = p[1]
            else:
                p[0] = 1.0

        def p_signed_int(p):
            """
            signed_int : SIGN UINT
            """
            p[0] = p[1] * p[2]

        def p_signed_float(p):
            """
            signed_float : sign UINT
                         | sign UFLOAT
            """
            p[0] = p[1] * p[2]

        def p_error(p):
            raise ValueError()

        return parsing.yacc(tabmodule="ogip_parsetab", package="astropy/units")

    @classmethod
    def _validate_unit(cls, unit, detailed_exception=True):
        if unit not in cls._units:
            if detailed_exception:
                raise ValueError(
                    f"Unit '{unit}' not supported by the OGIP standard. "
                    + utils.did_you_mean_units(
                        unit,
                        cls._units,
                        cls._deprecated_units,
                        cls._to_decomposed_alternative,
                    ),
                )
            else:
                raise ValueError()

        if unit in cls._deprecated_units:
            utils.unit_deprecation_warning(
                unit, cls._units[unit], "OGIP", cls._to_decomposed_alternative
            )

    @classmethod
    def _parse_unit(cls, unit, detailed_exception=True):
        cls._validate_unit(unit, detailed_exception=detailed_exception)
        return cls._units[unit]

    @classmethod
    def parse(cls, s, debug=False):
        s = s.strip()
        try:
            # This is a short circuit for the case where the string is
            # just a single unit name
            return cls._parse_unit(s, detailed_exception=False)
        except ValueError:
            try:
                return core.Unit(cls._parser.parse(s, lexer=cls._lexer, debug=debug))
            except ValueError as e:
                if str(e):
                    raise
                else:
                    raise ValueError(f"Syntax error parsing unit '{s}'")

    @classmethod
    def _get_unit_name(cls, unit):
        name = super()._get_unit_name(unit)
        cls._validate_unit(name)
        return name

    @classmethod
    def _format_superscript(cls, number):
        return f"**({number})" if "/" in number else f"**{number}"

    @classmethod
    def to_string(cls, unit, fraction="inline"):
        # Remove units that aren't known to the format
        unit = utils.decompose_to_known_units(unit, cls._get_unit_name)

        if isinstance(unit, core.CompositeUnit):
            # Can't use np.log10 here, because p[0] may be a Python long.
            if math.log10(unit.scale) % 1.0 != 0.0:
                warnings.warn(
                    f"'{unit.scale}' scale should be a power of 10 in OGIP format",
                    core.UnitsWarning,
                )

        return super().to_string(unit, fraction=fraction)

    @classmethod
    def _to_decomposed_alternative(cls, unit):
        # Remove units that aren't known to the format
        unit = utils.decompose_to_known_units(unit, cls._get_unit_name)

        if isinstance(unit, core.CompositeUnit):
            # Can't use np.log10 here, because p[0] may be a Python long.
            if math.log10(unit.scale) % 1.0 != 0.0:
                scale = unit.scale
                unit = copy.copy(unit)
                unit._scale = 1.0
                return (
                    f"{generic._to_string(cls, unit)} (with data multiplied by {scale})"
                )

        return super().to_string(unit)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from astropy.coordinates.baseframe import frame_transform_graph
from astropy.coordinates.matrix_utilities import matrix_transpose, rotation_matrix
from astropy.coordinates.transformations import StaticMatrixTransform

from .galactic import Galactic
from .supergalactic import Supergalactic


@frame_transform_graph.transform(StaticMatrixTransform, Galactic, Supergalactic)
def gal_to_supergal():
    return (
        rotation_matrix(90, "z")
        @ rotation_matrix(90 - Supergalactic._nsgp_gal.b.degree, "y")
        @ rotation_matrix(Supergalactic._nsgp_gal.l.degree, "z")
    )


@frame_transform_graph.transform(StaticMatrixTransform, Supergalactic, Galactic)
def supergal_to_gal():
    return matrix_transpose(gal_to_supergal())



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numbers

import numpy as np

from astropy.units import (
    CompositeUnit,
    Unit,
    UnitConversionError,
    UnitsError,
    UnitTypeError,
    dimensionless_unscaled,
    photometric,
)
from astropy.utils.compat.numpycompat import NUMPY_LT_2_0

from .core import FunctionQuantity, FunctionUnitBase
from .units import dB, dex, mag

__all__ = [
    "LogUnit",
    "MagUnit",
    "DexUnit",
    "DecibelUnit",
    "LogQuantity",
    "Magnitude",
    "Decibel",
    "Dex",
    "STmag",
    "ABmag",
    "M_bol",
    "m_bol",
]


class LogUnit(FunctionUnitBase):
    """Logarithmic unit containing a physical one.

    Usually, logarithmic units are instantiated via specific subclasses
    such `~astropy.units.MagUnit`, `~astropy.units.DecibelUnit`, and
    `~astropy.units.DexUnit`.

    Parameters
    ----------
    physical_unit : `~astropy.units.Unit` or `string`
        Unit that is encapsulated within the logarithmic function unit.
        If not given, dimensionless.

    function_unit :  `~astropy.units.Unit` or `string`
        By default, the same as the logarithmic unit set by the subclass.

    """

    # the four essential overrides of FunctionUnitBase
    @property
    def _default_function_unit(self):
        return dex

    @property
    def _quantity_class(self):
        return LogQuantity

    def from_physical(self, x):
        """Transformation from value in physical to value in logarithmic units.
        Used in equivalency.
        """
        return dex.to(self._function_unit, np.log10(x))

    def to_physical(self, x):
        """Transformation from value in logarithmic to value in physical units.
        Used in equivalency.
        """
        return 10 ** self._function_unit.to(dex, x)

    # ^^^^ the four essential overrides of FunctionUnitBase

    # add addition and subtraction, which imply multiplication/division of
    # the underlying physical units
    def _add_and_adjust_physical_unit(self, other, sign_self, sign_other):
        """Add/subtract LogUnit to/from another unit, and adjust physical unit.

        self and other are multiplied by sign_self and sign_other, resp.

        We wish to do:   ±lu_1 + ±lu_2  -> lu_f          (lu=logarithmic unit)
                  and     pu_1^(±1) * pu_2^(±1) -> pu_f  (pu=physical unit)

        Raises
        ------
        UnitsError
            If function units are not equivalent.
        """
        # First, insist on compatible logarithmic type. Here, plain u.mag,
        # u.dex, and u.dB are OK, i.e., other does not have to be LogUnit
        # (this will indirectly test whether other is a unit at all).
        try:
            getattr(other, "function_unit", other)._to(self._function_unit)
        except AttributeError:
            # if other is not a unit (i.e., does not have _to).
            return NotImplemented
        except UnitsError:
            raise UnitsError(
                "Can only add/subtract logarithmic units of compatible type."
            )

        other_physical_unit = getattr(other, "physical_unit", dimensionless_unscaled)
        physical_unit = CompositeUnit(
            1, [self._physical_unit, other_physical_unit], [sign_self, sign_other]
        )

        return self._copy(physical_unit)

    def __neg__(self):
        return self._copy(self.physical_unit ** (-1))

    def __add__(self, other):
        # Only know how to add to a logarithmic unit with compatible type,
        # be it a plain one (u.mag, etc.,) or another LogUnit
        return self._add_and_adjust_physical_unit(other, +1, +1)

    def __radd__(self, other):
        return self._add_and_adjust_physical_unit(other, +1, +1)

    def __sub__(self, other):
        return self._add_and_adjust_physical_unit(other, +1, -1)

    def __rsub__(self, other):
        # here, in normal usage other cannot be LogUnit; only equivalent one
        # would be u.mag,u.dB,u.dex.  But might as well use common routine.
        return self._add_and_adjust_physical_unit(other, -1, +1)


class MagUnit(LogUnit):
    """Logarithmic physical units expressed in magnitudes.

    Parameters
    ----------
    physical_unit : `~astropy.units.Unit` or `string`
        Unit that is encapsulated within the magnitude function unit.
        If not given, dimensionless.

    function_unit :  `~astropy.units.Unit` or `string`
        By default, this is ``mag``, but this allows one to use an equivalent
        unit such as ``2 mag``.
    """

    @property
    def _default_function_unit(self):
        return mag

    @property
    def _quantity_class(self):
        return Magnitude


class DexUnit(LogUnit):
    """Logarithmic physical units expressed in magnitudes.

    Parameters
    ----------
    physical_unit : `~astropy.units.Unit` or `string`
        Unit that is encapsulated within the magnitude function unit.
        If not given, dimensionless.

    function_unit :  `~astropy.units.Unit` or `string`
        By default, this is ``dex``, but this allows one to use an equivalent
        unit such as ``0.5 dex``.
    """

    @property
    def _default_function_unit(self):
        return dex

    @property
    def _quantity_class(self):
        return Dex

    def to_string(self, format="generic"):
        if format == "cds":
            if self.physical_unit == dimensionless_unscaled:
                return "[-]"  # by default, would get "[---]".
            else:
                return f"[{self.physical_unit.to_string(format=format)}]"
        else:
            return super().to_string()


class DecibelUnit(LogUnit):
    """Logarithmic physical units expressed in dB.

    Parameters
    ----------
    physical_unit : `~astropy.units.Unit` or `string`
        Unit that is encapsulated within the decibel function unit.
        If not given, dimensionless.

    function_unit :  `~astropy.units.Unit` or `string`
        By default, this is ``dB``, but this allows one to use an equivalent
        unit such as ``2 dB``.
    """

    @property
    def _default_function_unit(self):
        return dB

    @property
    def _quantity_class(self):
        return Decibel


class LogQuantity(FunctionQuantity):
    """A representation of a (scaled) logarithm of a number with a unit.

    Parameters
    ----------
    value : number, `~astropy.units.Quantity`, `~astropy.units.LogQuantity`, or sequence of quantity-like.
        The numerical value of the logarithmic quantity. If a number or
        a `~astropy.units.Quantity` with a logarithmic unit, it will be
        converted to ``unit`` and the physical unit will be inferred from
        ``unit``.  If a `~astropy.units.Quantity` with just a physical unit,
        it will converted to the logarithmic unit, after, if necessary,
        converting it to the physical unit inferred from ``unit``.

    unit : str, `~astropy.units.UnitBase`, or `~astropy.units.FunctionUnitBase`, optional
        For an `~astropy.units.FunctionUnitBase` instance, the
        physical unit will be taken from it; for other input, it will be
        inferred from ``value``. By default, ``unit`` is set by the subclass.

    dtype : `~numpy.dtype`, optional
        The ``dtype`` of the resulting Numpy array or scalar that will
        hold the value.  If not provided, is is determined automatically
        from the input value.

    copy : bool, optional
        If `True` (default), then the value is copied.  Otherwise, a copy will
        only be made if ``__array__`` returns a copy, if value is a nested
        sequence, or if a copy is needed to satisfy an explicitly given
        ``dtype``.  (The `False` option is intended mostly for internal use,
        to speed up initialization where a copy is known to have been made.
        Use with care.)

    Examples
    --------
    Typically, use is made of an `~astropy.units.FunctionQuantity`
    subclass, as in::

        >>> import astropy.units as u
        >>> u.Magnitude(-2.5)
        <Magnitude -2.5 mag>
        >>> u.Magnitude(10.*u.count/u.second)
        <Magnitude -2.5 mag(ct / s)>
        >>> u.Decibel(1.*u.W, u.DecibelUnit(u.mW))  # doctest: +FLOAT_CMP
        <Decibel 30. dB(mW)>

    """

    # only override of FunctionQuantity
    _unit_class = LogUnit

    # additions that work just for logarithmic units
    def __add__(self, other):
        # Add function units, thus multiplying physical units. If no unit is
        # given, assume dimensionless_unscaled; this will give the appropriate
        # exception in LogUnit.__add__.
        new_unit = self.unit + getattr(other, "unit", dimensionless_unscaled)
        # Add actual logarithmic values, rescaling, e.g., dB -> dex.
        result = self._function_view + getattr(other, "_function_view", other)
        return self._new_view(result, new_unit)

    def __radd__(self, other):
        return self.__add__(other)

    def __iadd__(self, other):
        new_unit = self.unit + getattr(other, "unit", dimensionless_unscaled)
        # Do calculation in-place using _function_view of array.
        function_view = self._function_view
        function_view += getattr(other, "_function_view", other)
        self._set_unit(new_unit)
        return self

    def __sub__(self, other):
        # Subtract function units, thus dividing physical units.
        new_unit = self.unit - getattr(other, "unit", dimensionless_unscaled)
        # Subtract actual logarithmic values, rescaling, e.g., dB -> dex.
        result = self._function_view - getattr(other, "_function_view", other)
        return self._new_view(result, new_unit)

    def __rsub__(self, other):
        new_unit = self.unit.__rsub__(getattr(other, "unit", dimensionless_unscaled))
        result = self._function_view.__rsub__(getattr(other, "_function_view", other))
        # Ensure the result is in right function unit scale
        # (with rsub, this does not have to be one's own).
        result = result.to(new_unit.function_unit)
        return self._new_view(result, new_unit)

    def __isub__(self, other):
        new_unit = self.unit - getattr(other, "unit", dimensionless_unscaled)
        # Do calculation in-place using _function_view of array.
        function_view = self._function_view
        function_view -= getattr(other, "_function_view", other)
        self._set_unit(new_unit)
        return self

    def __mul__(self, other):
        # Multiply by a float or a dimensionless quantity
        if isinstance(other, numbers.Number):
            # Multiplying a log means putting the factor into the exponent
            # of the unit
            new_physical_unit = self.unit.physical_unit**other
            result = self.view(np.ndarray) * other
            return self._new_view(result, self.unit._copy(new_physical_unit))
        else:
            return super().__mul__(other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __imul__(self, other):
        if isinstance(other, numbers.Number):
            new_physical_unit = self.unit.physical_unit**other
            function_view = self._function_view
            function_view *= other
            self._set_unit(self.unit._copy(new_physical_unit))
            return self
        else:
            return super().__imul__(other)

    def __truediv__(self, other):
        # Divide by a float or a dimensionless quantity
        if isinstance(other, numbers.Number):
            # Dividing a log means putting the denominator into the exponent
            # of the unit
            new_physical_unit = self.unit.physical_unit ** (1 / other)
            result = self.view(np.ndarray) / other
            return self._new_view(result, self.unit._copy(new_physical_unit))
        else:
            return super().__truediv__(other)

    def __itruediv__(self, other):
        if isinstance(other, numbers.Number):
            new_physical_unit = self.unit.physical_unit ** (1 / other)
            function_view = self._function_view
            function_view /= other
            self._set_unit(self.unit._copy(new_physical_unit))
            return self
        else:
            return super().__itruediv__(other)

    def __pow__(self, other):
        # We check if this power is OK by applying it first to the unit.
        try:
            other = float(other)
        except TypeError:
            return NotImplemented
        new_unit = self.unit**other
        new_value = self.view(np.ndarray) ** other
        return self._new_view(new_value, new_unit)

    def __ilshift__(self, other):
        try:
            other = Unit(other)
        except UnitTypeError:
            return NotImplemented

        if not isinstance(other, self._unit_class):
            return NotImplemented

        try:
            factor = self.unit.physical_unit._to(other.physical_unit)
        except UnitConversionError:
            # Maybe via equivalencies?  Now we do make a temporary copy.
            try:
                value = self._to_value(other)
            except UnitConversionError:
                return NotImplemented

            self.view(np.ndarray)[...] = value
        else:
            self.view(np.ndarray)[...] += self.unit.from_physical(factor)

        self._set_unit(other)
        return self

    # Methods that do not work for function units generally but are OK for
    # logarithmic units as they imply differences and independence of
    # physical unit.
    def var(self, axis=None, dtype=None, out=None, ddof=0):
        unit = self.unit.function_unit**2
        return self._wrap_function(np.var, axis, dtype, out=out, ddof=ddof, unit=unit)

    def std(self, axis=None, dtype=None, out=None, ddof=0):
        unit = self.unit._copy(dimensionless_unscaled)
        return self._wrap_function(np.std, axis, dtype, out=out, ddof=ddof, unit=unit)

    if NUMPY_LT_2_0:

        def ptp(self, axis=None, out=None):
            unit = self.unit._copy(dimensionless_unscaled)
            return self._wrap_function(np.ptp, axis, out=out, unit=unit)

    else:

        def __array_function__(self, function, types, args, kwargs):
            # TODO: generalize this to all supported functions!
            if function is np.ptp:
                unit = self.unit._copy(dimensionless_unscaled)
                return self._wrap_function(np.ptp, *args[1:], unit=unit, **kwargs)
            else:
                return super().__array_function__(function, types, args, kwargs)

    def diff(self, n=1, axis=-1):
        unit = self.unit._copy(dimensionless_unscaled)
        return self._wrap_function(np.diff, n, axis, unit=unit)

    def ediff1d(self, to_end=None, to_begin=None):
        unit = self.unit._copy(dimensionless_unscaled)
        return self._wrap_function(np.ediff1d, to_end, to_begin, unit=unit)

    _supported_functions = FunctionQuantity._supported_functions | {
        getattr(np, function) for function in ("var", "std", "ptp", "diff", "ediff1d")
    }


class Dex(LogQuantity):
    _unit_class = DexUnit


class Decibel(LogQuantity):
    _unit_class = DecibelUnit


class Magnitude(LogQuantity):
    _unit_class = MagUnit


dex._function_unit_class = DexUnit
dB._function_unit_class = DecibelUnit
mag._function_unit_class = MagUnit


STmag = MagUnit(photometric.STflux)
STmag.__doc__ = "ST magnitude: STmag=-21.1 corresponds to 1 erg/s/cm2/A"

ABmag = MagUnit(photometric.ABflux)
ABmag.__doc__ = "AB magnitude: ABmag=-48.6 corresponds to 1 erg/s/cm2/Hz"

M_bol = MagUnit(photometric.Bol)
M_bol.__doc__ = (
    f"Absolute bolometric magnitude: M_bol=0 corresponds to L_bol0={photometric.Bol.si}"
)

m_bol = MagUnit(photometric.bol)
m_bol.__doc__ = (
    f"Apparent bolometric magnitude: m_bol=0 corresponds to f_bol0={photometric.bol.si}"
)
